__version__ = '1.0' # Also update setup.py

from .cram import (general_rat_func, nsolve_intervals, nsolve_points,
    CRAM_exp, get_CRAM_from_cache, CRAM_coeffs)

__all__ = ['general_rat_func', 'nsolve_intervals', 'nsolve_points',
    'CRAM_exp', 'get_CRAM_from_cache', 'CRAM_coeffs']

from .partialfrac import (t, thetas_alphas, thetas_alphas_to_expr_real,
    thetas_alphas_to_expr_complex, thetas_alphas_to_expr_complex2, allroots,
    multiply_vector, customre)

__all__ += ['t', 'thetas_alphas', 'thetas_alphas_to_expr_real',
    'thetas_alphas_to_expr_complex', 'thetas_alphas_to_expr_complex2',
    'allroots', 'multiply_vector', 'customre']

from .codegen import (MatrixNumPyPrinter, autoeye, numpy_solve_with_autoeye,
    scipy_sparse_solve_with_autoeye, scipy_translations,
    scipy_translations_autoeye, CRAM_matrix_exp_lambdify)

__all__ += ['MatrixNumPyPrinter', 'autoeye', 'numpy_solve_with_autoeye',
    'scipy_sparse_solve_with_autoeye', 'scipy_translations',
    'scipy_translations_autoeye', 'CRAM_matrix_exp_lambdify']

from .origen import (execute_origen, origen_to_array,
    origen_data_to_array_weighted, origen_data_to_array_materials,
    origen_data_to_array_atom_fraction, initial_vector, compute_mismatch)

__all__ += ['execute_origen', 'origen_to_array',
    'origen_data_to_array_weighted', 'origen_data_to_array_materials',
    'origen_data_to_array_atom_fraction', 'initial_vector',
    'compute_mismatch']

from .origen_all import (ALL_LIBS, INITIAL_NUCS, DAY, YEAR, MONTH, TIME_STEPS,
    PHI)

__all__ += ['ALL_LIBS', 'INITIAL_NUCS', 'DAY', 'YEAR', 'MONTH', 'TIME_STEPS', 'PHI']

from .gensolve import generate

__all__ += ['generate']

from .generate_json import common_mat, generate_json

__all__ += ['common_mat', 'generate_json']

from .tape9utils import (LN2, DECAY_RXS, PAROFF, PAROFFM, THRESHOLD,
    origen_to_name, decay_data, XS_RXS, XS_TO_ORIGEN, ORIGEN_TO_XS, find_nlb,
    cross_section_data, sort_nucs, create_dok, dok_to_sparse_info,
    find_decaylib, SPMAT_FORMATS, tape9_to_sparse, normalize_tape9s)

__all__ += ['LN2', 'DECAY_RXS', 'PAROFF', 'PAROFFM', 'THRESHOLD',
    'origen_to_name', 'decay_data', 'XS_RXS', 'XS_TO_ORIGEN', 'ORIGEN_TO_XS',
    'find_nlb', 'cross_section_data', 'sort_nucs', 'create_dok',
    'dok_to_sparse_info', 'find_decaylib', 'SPMAT_FORMATS', 'tape9_to_sparse',
    'normalize_tape9s']

from .util import (diff_strs, relative_error, mean_log10_relative_error,
    plot_in_terminal, plt_show_in_terminal, log_function_args,
    save_sparse_csr, load_sparse_csr, time_func, memoize)

__all__ += ['diff_strs', 'relative_error', 'mean_log10_relative_error',
    'plot_in_terminal', 'plt_show_in_terminal', 'log_function_args',
    'save_sparse_csr', 'load_sparse_csr', 'time_func', 'memoize']



from argparse import ArgumentParser
import json
import os

from jinja2 import Environment
from sympy import im

from .cram import get_CRAM_from_cache
from .partialfrac import thetas_alphas

HEADER = """\
/* This file was generated automatically with transmutagen. */
#ifndef {{namespace.upper()}}_SOLVE_C
#define {{namespace.upper()}}_SOLVE_C

#include <complex.h>

typedef struct {{namespace}}_info_tag {
  int n;
  int nnz;
  int* i;
  int* j;
  char** nucs;
} {{namespace}}_info_t;

extern {{namespace}}_info_t {{namespace}}_info;
{% if py_solve %}
{%- for type, typefuncname in types %}
void {{namespace}}_solve_{{typefuncname}}({{type}}* A, {{type}}* b, {{type}}* x);
void {{namespace}}_diag_add_{{typefuncname}}({{type}}* A, {{type}} alpha);
void {{namespace}}_dot_{{typefuncname}}({{type}}* A, {{type}}* x, {{type}}* y);
void {{namespace}}_scalar_times_vector_{{typefuncname}}({{type}}, {{type}}*);
{% endfor %}
{%- endif %}
void {{namespace}}_solve_special(double* A, double complex theta, double complex alpha, double* b, double complex* x);
{%- for degree in degrees %}
void {{namespace}}_expm_multiply{{degree}}(double* A, double* b, double* x);
{%- endfor %}
#endif

"""

SRC = """\
/* This file was generated automatically with transmutagen. */
#include <string.h>

#include "solve.h"

const int {{namespace.upper()}}_I[{{NNZ}}] =
  { {%- for i, j in sorted(ij) %}{{i}},{% endfor -%} };

const int {{namespace.upper()}}_J[{{NNZ}}] =
  { {%- for i, j in sorted(ij) %}{{j}},{% endfor -%} };

const char* {{namespace.upper()}}_NUCS[{{N}}] =
  { {%- for nuc in nucs %}"{{nuc}}",{% endfor -%} };

{{namespace}}_info_t {{namespace}}_info = {
  .n = {{N}},
  .nnz = {{NNZ}},
  .i = (int*) {{namespace.upper()}}_I,
  .j = (int*) {{namespace.upper()}}_J,
  .nucs = (char**) {{namespace.upper()}}_NUCS,
};

int {{namespace}}_ij(int i, int j) {
  int n = (i << 16) + j;
  switch (n) {
    {%- for i, j in sorted(ij) %}
    case {{(i * 2**16) + j}}:
        return {{ij[i, j]}};
    {%- endfor %}
    default:
        return -1;
  }
}

{%- if py_solve %}
{%- for type, typefuncname in types %}
void {{namespace}}_solve_{{typefuncname}}({{type}}* A, {{type}}* b, {{type}}* x) {
  /* Decompose first */
  {{type}} LU [{{NIJK}}];
  memcpy(LU, A, {{NNZ}}*sizeof({{type}}));
  memset(LU+{{NNZ}}, 0, {{NIJK-NNZ}}*sizeof({{type}}));
  {%- for i in range(N) %}
  {%- for j in range(i+1, N) %}
  {%- if (j, i) in ijk %}
  LU[{{ijk[j, i]}}] /= LU[{{ijk[i, i]}}];
  {%- for k in range(i+1, N) %}
  {%- if (i, k) in ijk %}
  LU[{{ijk[j, k]}}] -= LU[{{ijk[j, i]}}] * LU[{{ijk[i, k]}}];
  {%- endif %}
  {%- endfor %}
  {%- endif %}
  {%- endfor %}
  {%- endfor %}

  /* Perform Solve */
  memcpy(x, b, {{N}}*sizeof({{type}}));
  {%- for i in range(N) %}{% if more_than_fore[i] %}
  x[{{i}}] = x[{{i}}]{% for j in range(i) %}{%if (i, j) in ijk%} - LU[{{ijk[i, j]}}]*x[{{j}}]{%endif%}{% endfor %};
  {%- endif %}
  {%- endfor %}
  /* Backward calc */
  {%- for i in range(N-1, -1, -1) %}{%if more_than_back[i]%}
  x[{{i}}] = x[{{i}}]{% for j in range(i+1, N) %}{%if (i, j) in ijk%} - LU[{{ijk[i, j]}}]*x[{{j}}]{%endif%}{% endfor %};
  {%- endif %}
  x[{{i}}] /= LU[{{ijk[i, i]}}];
  {%- endfor %}
}

void {{namespace}}_diag_add_{{typefuncname}}({{type}}* A, {{type}} theta) {
  /* In-place, performs the addition A + theta I, for a scalar theta. */
  {%- for i in range(N) %}
  A[{{ij[i, i]}}] += theta;
  {%- endfor %}
}

void {{namespace}}_dot_{{typefuncname}}({{type}}* A, {{type}}* x, {{type}}* y) {
  /* Performs the caclulation Ax = y and returns y */
  {%- for i in range(N) %}
  y[{{i}}] ={% for j in range(N) %}{% if (i,j) in ij %} + A[{{ij[i, j]}}]*x[{{j}}]{% endif %}{% endfor %};
  {%- endfor %}
}

void {{namespace}}_scalar_times_vector_{{typefuncname}}({{type}} alpha, {{type}}* v) {
  /* In-place, performs alpha*v, for a scalar alpha and vector v. */
  {%- for i in range(N) %}
  v[{{i}}] *= alpha;
  {%- endfor %}
}

{%- endfor %}
{%- endif %}

void {{namespace}}_solve_special(double* A, double complex theta, double complex alpha, double* b, double complex* x) {
  /* Solves (A + theta*I)x = alpha*b and stores the result in x */
  double complex LU [{{NIJK}}];

  /* LU = A + theta*I */
  {%- for i in range(NNZ) %}
  {%- if i in diagonals %}
  LU[{{i}}] = theta + A[{{i}}];
  {%- else %}
  LU[{{i}}] = A[{{i}}];
  {%- endif %}
  {%- endfor %}

  memset(LU+{{NNZ}}, 0, {{NIJK-NNZ}}*sizeof(double complex));

  /* Decompose first */
  {%- for i in range(N) %}
  {%- for j in range(i+1, N) %}
  {%- if (j, i) in ijk %}
  LU[{{ijk[j, i]}}] /= LU[{{ijk[i, i]}}];
  {%- for k in range(i+1, N) %}
  {%- if (i, k) in ijk %}
  LU[{{ijk[j, k]}}] -= LU[{{ijk[j, i]}}] * LU[{{ijk[i, k]}}];
  {%- endif %}
  {%- endfor %}
  {%- endif %}
  {%- endfor %}
  {%- endfor %}

  /* Multiply x by alpha and perform Solve */
  {%- for i in range(N) %}
  x[{{i}}] = alpha*b[{{i}}]{% for j in range(i) %}{%if (i, j) in ijk%} - LU[{{ijk[i, j]}}]*x[{{j}}]{%endif%}{% endfor %};
  {%- endfor %}
  /* Backward calc */
  {%- for i in range(N-1, -1, -1) %}{%if more_than_back[i]%}
  x[{{i}}] = x[{{i}}]{% for j in range(i+1, N) %}{%if (i, j) in ijk%} - LU[{{ijk[i, j]}}]*x[{{j}}]{%endif%}{% endfor %};
  {%- endif %}
  x[{{i}}] /= LU[{{ijk[i, i]}}];
  {%- endfor %}
}

{% for degree in degrees %}
void {{namespace}}_expm_multiply{{degree}}(double* A, double* b, double* x) {
    /* Computes exp(A)*b and stores the result in x */
    {%- for i in range(degree//2) %}
    double complex x{{i}} [{{N}}];
    {%- endfor %}

    {% set thetas, alphas, alpha0 = get_thetas_alphas(degree) -%}
    {% for theta, alpha in sorted(zip(thetas, alphas), key=abs0) if im(theta) >= 0 %}
    {{namespace}}_solve_special(A, {{ -theta}}, {{2*alpha}}, b, x{{loop.index0}});
    {%- endfor %}

    {% for i in range(N) %}
    x[{{i}}] = (double)creal({%- for j in range(degree//2) %}+x{{j}}[{{i}}]{%- endfor %}) + {{alpha0}}*b[{{i}}];
    {%- endfor %}
}

{% endfor %}
"""

def make_ijk(ij, N):
    ijk = ij.copy()
    idx = len(ij)
    for i in range(N):
        for j in range(i+1, N):
            if (j, i) not in ijk:
                continue
            for k in range(i+1, N):
                if (i, k) in ijk and (j, k) not in ijk:
                    ijk[j, k] = idx
                    idx += 1
    return ijk

def get_thetas_alphas(degree, prec=200, use_cache=True):
    rat_func = get_CRAM_from_cache(degree, prec, log=True, plot=False, use_cache=use_cache)

    thetas, alphas, alpha0 = thetas_alphas(rat_func, prec)
    return thetas, alphas, alpha0

def generate(json_file=os.path.join(os.path.dirname(__file__), 'data/gensolve.json'),
    outfile=None, degrees=None, py_solve=False, namespace='transmutagen'):

    if degrees is None:
        degrees = [6, 8, 10, 12, 14, 16, 18] if py_solve else [14]
    if not outfile:
        outfile='py_solve/py_solve/solve.c' if py_solve else 'solve.c'
    # outfile should always end in .c
    headerfile = outfile[:-2] + '.h'

    with open(json_file) as f:
        json_data = json.load(f)

    nucs = json_data['nucs']
    N = len(nucs)
    ijkeys = [(nucs.index(j), nucs.index(i)) for i, j in json_data['tofrom']]
    ij = {k: l for l, k in enumerate(sorted(ijkeys))}
    ijk = make_ijk(ij, N)
    diagonals = {ij[i, i]: i for i in range(N)}
    more_than_fore = [len([j for j in range(i+1) if (i, j) in ijk]) > 1 for i in range(N)]
    more_than_back = [len([j for j in range(i, N) if (i, j) in ijk]) > 1 for i in range(N)]
    types = [  # C type, type function name
             ('double', 'double'),
             ('double complex', 'complex')]
    env = Environment()
    src_template = env.from_string(SRC, globals=globals())
    src = src_template.render(N=N, ij=ij, ijk=ijk, nucs=nucs, sorted=sorted,
        len=len, more_than_back=more_than_back, NNZ=len(ij), NIJK=len(ijk),
        more_than_fore=more_than_fore, types=types, namespace=namespace,
        diagonals=diagonals, degrees=degrees, py_solve=py_solve,
        get_thetas_alphas=get_thetas_alphas, im=im, abs0=lambda i:abs(i[0]),
        zip=zip, enumerate=enumerate)
    header_template = env.from_string(HEADER, globals=globals())
    header = header_template.render(types=types, degrees=degrees,
        py_solve=py_solve, namespace=namespace)
    print("Writing", outfile)
    with open(outfile, 'w') as f:
        f.write(src)
    print("Writing", headerfile)
    with open(headerfile, 'w') as f:
        f.write(header)
    # If this changes, also update py_solve/setup.py
    compiler_flags = ['-O0', '-fcx-fortran-rules', '-fcx-limited-range',
        '-ftree-sra', '-ftree-ter', '-fexpensive-optimizations']

    print("With gcc, it is recommended to compile the following flags:", ' '.join(compiler_flags))

def main(args=None):
    p = ArgumentParser('gensolver')
    p.add_argument('--json-file',
        default=os.path.join(os.path.dirname(__file__), 'data/gensolve.json'),
        help="""Location of the json input file. An input file can be
        generated from ORIGEN libraries with python -m
        transmutagen.generate_json. The default is %(default)r.""")
    p.add_argument('--py-solve', action='store_true', help="""Generate code for
        py_solve.""")
    p.add_argument('--degrees', nargs='+', default=None, help="""expm_multiply
        degrees to generate. The default is 14, unless --py-solve is
        specified, in which case the default is '6 8 10 12 14 16 18'. The
        orders should be even integers only.""", metavar="DEGREE")
    p.add_argument('--outfile', help="""Location to write the C file to.
        Should end in '.c'. The default is 'solve.c', unless --py-solve is
        specified, in which case the default is 'py_solve/py_solve/solve.c'.
        The header file will be generated alongside it.""")
    p.add_argument('--namespace', default='transmutagen', help="""Namespace
        for the generated solver. The default is %(default)r.""")

    ns = p.parse_args(args=args)
    if ns.outfile and not ns.outfile.endswith('.c'):
        p.error("--outfile should end with '.c'")
    arguments = ns.__dict__.copy()

    generate(**arguments)


if __name__ == "__main__":
    main()



"""Some utilities for dealing with Origen 2.2 TAPE9 files."""
import os
import warnings
from collections import defaultdict

import numpy as np

import scipy.sparse

from pyne import utils
utils.toggle_warnings()
warnings.simplefilter('ignore')
from pyne import data
from pyne import rxname
from pyne import nucname
from pyne.origen22 import parse_tape9

LN2 = np.log(2.0)
DECAY_RXS = ['bminus', 'bplus', 'ec', 'alpha', 'it', 'sf', 'bminus_n']
PAROFF = {
    'frac_beta_minus': 10000000,
    'frac_beta_minus_x': 10000001,
    'frac_beta_plus_or_electron_capture': -10000000,
    'frac_beta_plus_or_electron_capture_x': -9999999,
    'frac_alpha': -20040000,
    'frac_isomeric_transition': -1,
    #'frac_spont_fiss': 0,
    'frac_beta_n': 9990000,
    }
PAROFFM = {
    'frac_beta_minus': 9999999,
    'frac_beta_minus_x': 10000000,
    'frac_beta_plus_or_electron_capture': -10000001,
    'frac_beta_plus_or_electron_capture_x': -10000000,
    'frac_alpha': -20040001,
    'frac_isomeric_transition': -1,
    #'frac_spont_fiss': 0,
    'frac_beta_n': 9989999,
    }
# This threshold is a decay constant for a half-life that is 10x the age of
# the universe.
THRESHOLD = LN2/(10 * 13.7e9 * 365.25* 24 * 3600)


def origen_to_name(nuc):
    """Takes a nuclide in Origen format and returns it's human readable name."""
    return nucname.name(nucname.zzaaam_to_id(int(nuc)))


def decay_data(t9, nlb=(1, 2, 3), threshold=THRESHOLD, nucs=None):
    """Gets decay data from a TAPE9.

    Parameters
    ----------
    t9 : dict
        A TAPE9 dict that contains the decay data.
    nlb : tuple of ints, optional
        The decay library numbers. Usually this is just (1, 2, 3).
    threshold : float, optional
        A cutoff for when we consider nuclides stable or a decay mode
        forbidden. This is given as a decay constant and the default is
        for a half-life that is 10x the age of the universe.
    nucs : set or None, optional
        The known set of nuclides.

    Returns
    -------
    nucs : set
        The set of nuclide names.
    decay_consts : dict
        Mapping from nuclide names to decay constants [1/sec]
    gammas : dict
        Mapping from (i, j) nuclide name tuples to the branch ratio
        fraction. This is [unitless] and the sum over all j for a given
        i is guarenteed to sum to 1.
    """
    nucs = set() if nucs is None else nucs
    decay_consts = {}
    gammas = {}
    for n in nlb:
        decay_consts.update({origen_to_name(nuc): LN2/val
                             for nuc, val in t9[n]['half_life'].items()
                             if nuc != 162500})
        for key in t9[n]:
            if not key.startswith('frac_') or key == 'frac_natural_abund' or  \
                                              key == 'frac_spont_fiss':
                continue
            for nuc, val in t9[n][key].items():
                if val < threshold:
                    continue  # forbidden decay
                nname = origen_to_name(nuc)
                if decay_consts[nname] < threshold:
                    continue  # stable nuclide
                poff = PAROFF if int(nuc)%10 == 0 else PAROFFM
                child = nucname.zzaaam_to_id(int(nuc)) + poff[key]
                cname = nucname.name(child)
                nucs.add(nname)
                nucs.add(cname)
                gammas[nname, cname] = val
    # add β- decays
    for nname, val in decay_consts.items():
        if val < threshold:
            continue  # stable nuclide
        gamma_total = sum([v for (n, c), v in gammas.items() if n == nname])
        if gamma_total < 1.0:
            parid = nucname.id(nname)
            poff = PAROFF if parid%10 == 0 else PAROFFM
            child = parid + poff['frac_beta_minus']
            cname = nucname.name(child)
            gammas[nname, cname] = 1.0 - gamma_total
            gamma_total = 1.0
            nucs.add(nname)
            nucs.add(cname)
        if gamma_total > 1.0:
            biggest_gamma = max([(i, j) for i, j in gammas if i == nname],
                                key=lambda t: gammas[t])
            gammas[biggest_gamma] = gammas[biggest_gamma] + 1 - gamma_total
    return nucs, decay_consts, gammas


XS_RXS = ['gamma', 'z_2n', 'z_3n', 'alpha', 'fission', 'proton', 'gamma_1', 'z_2n_1']
XS_TO_ORIGEN = {'gamma': 'gamma', 'z_2n': '2n', 'z_3n': '3n', 'alpha': 'alpha',
                'fission': 'f', 'proton': 'p', 'gamma_1': 'gamma_x', 'z_2n_1': '2n_x'}
ORIGEN_TO_XS = {v: k for k, v in XS_TO_ORIGEN.items()}


def find_nlb(t9, nlb=None):
    """Finds the library numbers in the TAPE9 file."""
    if nlb is not None:
        return nlb
    nlb = set(t9.keys()) - {1, 2, 3}
    if len(nlb) > 3:
        raise ValueError("Too many libraries found in TAPE9 file.")
    return tuple(sorted(nlb))


def cross_section_data(t9, nlb=None, threshold=THRESHOLD, nucs=None):
    """Gets decay data from a TAPE9.

    Parameters
    ----------
    t9 : dict
        A TAPE9 dict that contains the decay data.
    nlb : tuple of ints or None, optional
        The cross section library numbers. If None, this will attempt to discover
        the numbers in the library.
    threshold : float, optional
        A cutoff for when we consider nuclides stable or a decay mode
        forbidden. This is given as a decay constant and the default is
        for a half-life that is 10x the age of the universe.
    nucs : set or None, optional
        The known set of nuclides.

    Returns
    -------
    nucs : set
        The set of nuclide names.
    sigma_ij : dict
        Mapping from (i, j) nuclide name tuples to the cross section for this
        reaction. Note that this does not include the fission cross section [barns].
    sigma_fission : dict
        Mapping from fissionable nuclide names to the fission cross section [barns].
    fission_product_yields : dict
        Mapping from (i, j) nuclide name tuples to the fission product yields for
        that nuclide.
    """
    nlb = find_nlb(t9, nlb=nlb)
    nucs = set() if nucs is None else nucs
    sigma_ij = {}
    sigma_fission = {}
    fission_product_yields = {}
    for n in nlb:
        # grab sigma_ij cross sections
        for rx in t9[n]:
            if not rx.startswith('sigma_') or rx == 'sigma_f':
                continue
            _, _, orx = rx.partition('_')
            xs_rs = ORIGEN_TO_XS[orx]
            for nuc in t9[n][rx]:
                val = t9[n][rx][nuc]
                if val < threshold:
                    continue
                nname = nucname.name(int(nuc))
                try:
                    child = nucname.name(rxname.child(nname, xs_rs))
                except RuntimeError:
                    continue
                nucs.add(nname)
                nucs.add(child)
                sigma_ij[nname, child] = val
        # grab the fission cross section
        if 'sigma_f' in t9[n]:
            rx = 'sigma_f'
            for nuc in t9[n][rx]:
                val = t9[n][rx][nuc]
                if val < threshold:
                    continue
                nname = nucname.name(int(nuc))
                nucs.add(nname)
                sigma_fission[nname] = val
        # grab the fission product yields
        for rx in t9[n]:
            if not rx.endswith('_fiss_yield'):
                continue
            fromnuc, *_ = rx.partition('_')
            fromnuc = nucname.name(fromnuc)
            for k, v in t9[n][rx].items():
                if v < threshold:
                    continue
                tonuc = nucname.name(nucname.zzaaam_to_id(int(k)))
                # origen yields are in percent
                nucs.add(fromnuc)
                nucs.add(tonuc)
                fission_product_yields[fromnuc, tonuc] = v / 100
    return nucs, sigma_ij, sigma_fission, fission_product_yields


def sort_nucs(nucs):
    """Returns the canonical ordering of a collection of nulcides."""
    return sorted(nucs, key=nucname.id)


def create_dok(phi, nucs, decay_consts, gammas, sigma_ij, sigma_fission,
               fission_product_yields):
    """Creates a dictionary-of-keys representation of the transumation data.

    Parameters
    ----------
    phi : float
        The neutron flux in [n / cm^2 / sec]
    nucs : list
        The list of nuclide names in canonical order.
    decay_consts : dict
        Mapping from nuclide names to decay constants [1/sec]
    gammas : dict
        Mapping from (i, j) nuclide name tuples to the branch ratio
        fraction. This is [unitless] and the sum over all j for a given
        i is guarenteed to sum to 1.
    sigma_ij : dict
        Mapping from (i, j) nuclide name tuples to the cross section for this
        reaction. Note that this does not include the fission cross section [barns].
    sigma_fission : dict
        Mapping from fissionable nuclide names to the fission cross section [barns].
    fission_product_yields : dict
        Mapping from (i, j) nuclide name tuples to the fission product yields for
        that nuclide.

    Returns
    -------
    dok : defaultdict
        Mapping from (i, j) nuclide name tuples to the transmutation rate.
    """
    phi = phi * 1e-24  # flux, n / barn / s
    dok = defaultdict(float) # indexed by nuclide name
    # let's first add the cross section channels
    for i, j in sigma_ij:
        v = sigma_ij.get((i, j), 0.0) * phi
        dok[j, i] += v
        dok[i, i] -= v
    # now let's add the fission products
    for (i, j), fpy in fission_product_yields.items():
        dok[j, i] += fpy * sigma_fission.get(i, 0.0) * phi
    for i, sigf in sigma_fission.items():
        dok[i, i] -= sigf * phi
    # now let's add the decay consts
    for (i, j), g in gammas.items():
        dok[j, i] += g * decay_consts[i]
    for i, v in decay_consts.items():
        dok[i, i] -= v
    return dok


def dok_to_sparse_info(nucs, dok):
    """Converts the dict of keys to a sparse info

    Parameters
    ----------
    nucs : list
        The list of nuclide names in canonical order.
    dok : dict
        Dictionary of keys representing a sparse transmutation matrix.

    Returns
    -------
    rows : list
        Row indexs
    cols : list
        Column indexes
    vals : list
        Values at cooresponding row/col index
    shape : 2-tuple of ints
        Represents the size of the matirx
    """
    shape = (len(nucs), len(nucs))
    nuc_idx = {n: i for i, n in enumerate(nucs)}
    rows = []
    cols = []
    vals = []
    for (i, j), v in dok.items():
        if (v == 0.0) or (i not in nuc_idx) or (j not in nuc_idx):
            continue
        rows.append(nuc_idx[i])
        cols.append(nuc_idx[j])
        vals.append(v)
    return rows, cols, vals, shape


def find_decaylib(t9, tape9, decaylib):
    """Finds and loads a decay lib."""
    if len({1, 2, 3} & set(t9.keys())) == 3:
        return t9
    if not os.path.isabs(decaylib):
        d = os.path.dirname(tape9)
        decaylib = os.path.join(d, decaylib)
    decay = parse_tape9(decaylib)
    return decay


# not all sparse matrices accept row, col, val, shape data
SPMAT_FORMATS = {
    'bsr': scipy.sparse.bsr_matrix,
    'coo': scipy.sparse.coo_matrix,
    'csc': scipy.sparse.csc_matrix,
    'csr': scipy.sparse.csr_matrix,
    }


def tape9_to_sparse(tape9s, phi, format='csr', decaylib='decay.lib',
                    include_fission=True, threshold=THRESHOLD):
    """Converts a TAPE9 file to a sparse matrix.

    Parameters
    ----------
    tape9s : str or list of str
        The filename(s) of the tape file(s).
    phi : float
        The neutron flux in [n / cm^2 / sec]
    format : str, optional
        Format of the sparse matrix created.
    decaylib : str, optional
        A path to TAPE9 containg the decay data libraries if the decay libraries
        are not present in tape9. If this is a relative path, it is taken
        relative to the given tape9 location.
    include_fission : bool
        Flag for whether or not the fission data should be included in the
        resultant matrix.
    threshold : float, optional
        A cutoff for when we consider nuclides stable or a decay mode
        forbidden. This is given as a decay constant and the default is
        for a half-life that is 10x the age of the universe.

    Returns
    -------
    mat : scipy.sparse matrix
        A sparse matrix in the specified layout.
    nucs : list
        The list of nuclide names in canonical order.
    """
    all_decays_consts, all_gammas, all_sigma_ij, all_sigma_fission, all_fission_product_yields = [], [], [], [], []
    nucs = set()
    mats = []
    # seed initial nucs with known atomic masses
    data.atomic_mass('u235')
    for tape9 in tape9s:
        # print("Getting data for", tape9)
        t9 = parse_tape9(tape9)
        decay = find_decaylib(t9, tape9, decaylib)

        for i in data.atomic_mass_map.keys():
            if nucname.iselement(i):
                continue
            try:
                nucs.add(nucname.name(i))
            except RuntimeError:
                pass

        # get the tape 9 data
        nucs, decays_consts, gammas = decay_data(decay, threshold=threshold, nucs=nucs)
        nucs, sigma_ij, sigma_fission, fission_product_yields = cross_section_data(t9,
                                                                    threshold=threshold,
                                                                    nucs=nucs)

        if not include_fission:
            sigma_fission = {}
            fission_product_yields = {}
        all_decays_consts.append(decays_consts)
        all_gammas.append(gammas)
        all_sigma_ij.append(sigma_ij)
        all_sigma_fission.append(sigma_fission)
        all_fission_product_yields.append(fission_product_yields)


    nucs = sort_nucs(nucs)
    for i in range(len(tape9s)):
        dok = create_dok(phi, nucs, all_decays_consts[i], all_gammas[i], all_sigma_ij[i], all_sigma_fission[i],
                     all_fission_product_yields[i])
        rows, cols, vals, shape = dok_to_sparse_info(nucs, dok)
        mats.append(SPMAT_FORMATS[format]((vals, (rows, cols)), shape=shape))
    return mats, nucs


def normalize_tape9s(tape9s):
    from .origen_all import ALL_LIBS

    if isinstance(tape9s, str):
        tape9s = [tape9s]

    _tape9s = []
    for tape9 in tape9s[:]:
        if os.path.isdir(tape9):
            _tape9s.extend([os.path.join(tape9, i) for i in ALL_LIBS])
        else:
            _tape9s.append(tape9)
    return _tape9s



# PYTHON_ARGCOMPLETE_OK

import argparse
import os
from subprocess import run
import logging
from itertools import combinations

import numpy as np
from scipy.sparse import csr_matrix
import tables

from pyne.utils import toggle_warnings
import warnings
toggle_warnings()
warnings.simplefilter('ignore')

import pyne.data
import pyne.material
from pyne.origen22 import (nlbs, write_tape5_irradiation, write_tape4,
    parse_tape9, merge_tape9, write_tape9, parse_tape6)
from pyne.material import from_atom_frac

from .util import load_sparse_csr, time_func
from .tape9utils import origen_to_name
from .codegen import CRAM_matrix_exp_lambdify

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
# Change to WARN for less output
logger.setLevel(logging.INFO)

# ORIGEN = '/home/origen22/code/o2_therm_linux.exe'
ORIGEN = '/home/o2prec/o2prec'
decay_TAPE9 = "/home/origen22/libs/decay.lib"
LIBS_DIR = "/home/origen22/libs"
DATA_DIR = os.path.abspath(os.path.join(__file__, os.path.pardir,
    os.path.pardir, 'data'))

NUCLIDE_KEYS = ['activation_products', 'actinides', 'fission_products']

def execute_origen(xs_tape9, time, nuclide, phi, origen, decay_tape9):
    xs_tape9 = xs_tape9
    if not os.path.isabs(xs_tape9):
        xs_tape9 = os.path.join(LIBS_DIR, xs_tape9)

    parsed_xs_tape9 = parse_tape9(xs_tape9)
    parsed_decay_tape9 = parse_tape9(decay_tape9)

    merged_tape9 = merge_tape9([parsed_decay_tape9, parsed_xs_tape9])

    # Can set outfile to change directory, but the file name needs to be
    # TAPE9.INP.
    write_tape9(merged_tape9)

    decay_nlb, xsfpy_nlb = nlbs(parsed_xs_tape9)

    # Can set outfile, but the file name should be called TAPE5.INP.
    write_tape5_irradiation("IRF", time/(60*60*24), phi,
        xsfpy_nlb=xsfpy_nlb, cut_off=0, out_table_num=[4, 5],
        out_table_nes=[True, False, False])

    M = from_atom_frac({nuclide: 1}, mass=1, atoms_per_molecule=1)

    write_tape4(M)

    origen_time, data = time_func(run, origen)

    # Make pyne use naive atomic mass numbers to match ORIGEN
    for i in pyne.data.atomic_mass_map:
        pyne.data.atomic_mass_map[i] = float(pyne.nucname.anum(i))

    data = parse_tape6()

    logger.info("ORIGEN runtime: %s", origen_time)
    return origen_time, data

def load_data(datafile):
    import pyne.data
    # Make pyne use naive atomic mass numbers to match ORIGEN
    for i in pyne.data.atomic_mass_map:
        pyne.data.atomic_mass_map[i] = float(pyne.nucname.anum(i))

    with open(datafile) as f:
        return eval(f.read(), {'array': np.array, 'pyne': pyne})

def origen_to_array(origen_dict, nucs):
    new_data = np.zeros((len(nucs), 1))
    nuc_to_idx = {v: i for i, v in enumerate(nucs)}
    for i in origen_dict:
        new_data[nuc_to_idx[origen_to_name(i)]] += origen_dict[i][1]

    return new_data

def origen_data_to_array_weighted(ORIGEN_data, nucs, n_fission_fragments=2.004):
    # Table 5 is grams
    table_5_weights = {}
    table_5_nuclide = ORIGEN_data['table_5']['nuclide']
    for key in NUCLIDE_KEYS:
        table_5_weights[key] = np.sum(origen_to_array(table_5_nuclide[key], nucs), axis=0)
    table_5_weights['fission_products'] *= n_fission_fragments

    # Table 4 is atom fraction
    table_4_nuclide = ORIGEN_data['table_4']['nuclide']
    new_data = np.zeros((len(nucs), 1))
    for key in NUCLIDE_KEYS:
        new_data += table_5_weights[key]*origen_to_array(table_4_nuclide[key], nucs)

    return new_data

def origen_data_to_array_atom_fraction(ORIGEN_data, nucs):
    # Table 4 is atom fraction
    table_4_nuclide = ORIGEN_data['table_4']['nuclide']
    new_data = np.zeros((len(nucs), 1))
    for key in NUCLIDE_KEYS:
        new_data += origen_to_array(table_4_nuclide[key], nucs)

    return new_data

def origen_data_to_array_materials(ORIGEN_data, nucs):
    material = ORIGEN_data['materials'][1]
    new_data = np.zeros((len(nucs), 1))
    nuc_to_idx = {v: i for i, v in enumerate(nucs)}

    for nuc, atom_frac in material.to_atom_frac().items():
        new_data[nuc_to_idx[pyne.nucname.name(nuc)]] = atom_frac

    return new_data

def hash_data(vec, library, time, phi, n_fission_fragments):
    return hash((tuple(vec.flat), library, time, phi, n_fission_fragments))

def initial_vector(start_nuclide, nucs):
    nuc_to_idx = {v: i for i, v in enumerate(nucs)}
    return csr_matrix(([1], [[nuc_to_idx[start_nuclide]], [0]]),
        shape=(len(nucs), 1)).toarray()

def test_origen_data_sanity(ORIGEN_data):
    for table in ['table_4', 'table_5']:
        assert table in ORIGEN_data, table
        assert 'nuclide' in ORIGEN_data[table]

        nuclide = ORIGEN_data['table_4']['nuclide']

        # Sanity check
        for comb in combinations(NUCLIDE_KEYS, 2):
            a, b = comb
            for common in set.intersection(set(nuclide[a]), set(nuclide[b])):
                array_a, array_b = nuclide[a][common], nuclide[b][common]
                assert np.allclose(array_a, 0) \
                    or np.allclose(array_b, 0)
                    # or np.allclose(array_a, array_b)

def create_hdf5_table(file, lib, nucs):
    nucs_size = len(nucs)
    desc_common = [
        ('hash', np.int64),
        ('library', 'S8'),
        ('initial vector', np.float64, (nucs_size, 1)),
        ('time', np.float64),
        ('phi', np.float64),
        ('n_fission_fragments', np.float64),
    ]
    desc_origen = [
        ('execution time ORIGEN', np.float64),
        ('ORIGEN atom fraction', np.float64, (nucs_size, 1)),
        ('ORIGEN mass fraction', np.float64, (nucs_size, 1)),
        ]
    desc_cram_lambdify = [
        ('execution time CRAM lambdify', np.float64),
        ('CRAM lambdify atom fraction', np.float64, (nucs_size, 1)),
        ('CRAM lambdify mass fraction', np.float64, (nucs_size, 1)),
        ]
    desc_cram_py_solve = [
        ('execution time CRAM py_solve', np.float64),
        ('CRAM py_solve atom fraction', np.float64, (nucs_size, 1)),
        ('CRAM py_solve mass fraction', np.float64, (nucs_size, 1)),
        ]

    h5file = tables.open_file(file, mode="a", title="CRAM/ORIGEN test run data", filters=tables.Filters(complevel=1))
    h5file.create_group('/', lib, '%s data' % lib)
    h5file.create_table('/' + lib, 'origen', np.dtype(desc_common + desc_origen))
    h5file.create_table('/' + lib, 'cram-lambdify', np.dtype(desc_common + desc_cram_lambdify))
    h5file.create_table('/' + lib, 'cram-py_solve', np.dtype(desc_common + desc_cram_py_solve))
    h5file.create_array('/' + lib, 'nucs', np.array(nucs, 'S6'))

def save_file_origen(file, *, ORIGEN_data, lib, nucs, start_nuclide, time,
    phi, ORIGEN_time, n_fission_fragments=2.004):

    with tables.open_file(file, mode="a", title="ORIGEN and CRAM data",
        filters=tables.Filters(complevel=1)) as h5file:

        if lib not in h5file.root:
            create_hdf5_table(file, lib, nucs)

        table = h5file.get_node(h5file.root, lib + '/origen')
        table.row['initial vector'] = vec = initial_vector(start_nuclide, nucs)
        table.row['library'] = lib
        table.row['hash'] = hash_data(vec, lib, time, phi, n_fission_fragments)
        table.row['time'] = time
        table.row['phi'] = phi
        table.row['n_fission_fragments'] = n_fission_fragments
        table.row['execution time ORIGEN'] = ORIGEN_time
        table.row['ORIGEN atom fraction'] = origen_data_to_array_weighted(ORIGEN_data, nucs, n_fission_fragments=n_fission_fragments)
        table.row['ORIGEN mass fraction'] = origen_data_to_array_materials(ORIGEN_data, nucs)
        table.row.append()
        table.flush()

def save_file_cram_lambdify(file, *, CRAM_lambdify_res, lib, nucs, start_nuclide, time,
    phi, CRAM_lambdify_time, n_fission_fragments=2.004):
    assert len(CRAM_lambdify_res) == len(nucs)
    with tables.open_file(file, mode="a", title="ORIGEN and CRAM data",
        filters=tables.Filters(complevel=1)) as h5file:

        if lib not in h5file.root:
            create_hdf5_table(file, lib, nucs)

        table = h5file.get_node(h5file.root, lib + '/cram-lambdify')
        table.row['initial vector'] = vec = initial_vector(start_nuclide, nucs)
        table.row['library'] = lib
        table.row['hash'] = hash_data(vec, lib, time, phi, n_fission_fragments)
        table.row['time'] = time
        table.row['phi'] = phi
        table.row['n_fission_fragments'] = n_fission_fragments
        table.row['execution time CRAM lambdify'] = CRAM_lambdify_time
        table.row['CRAM lambdify atom fraction'] = CRAM_lambdify_res
        CRAM_lambdify_res_normalized = CRAM_lambdify_res/np.sum(CRAM_lambdify_res)
        table.row['CRAM lambdify mass fraction'] = CRAM_lambdify_res_normalized
        table.row.append()
        table.flush()

def save_file_cram_py_solve(file, *, CRAM_py_solve_res, lib, nucs, start_nuclide, time,
    phi, CRAM_py_solve_time, n_fission_fragments=2.004):
    assert len(CRAM_py_solve_res) == len(nucs)
    with tables.open_file(file, mode="a", title="ORIGEN and CRAM data",
        filters=tables.Filters(complevel=1)) as h5file:

        if lib not in h5file.root:
            create_hdf5_table(file, lib, nucs)

        table = h5file.get_node(h5file.root, lib + '/cram-py_solve')
        table.row['initial vector'] = vec = initial_vector(start_nuclide, nucs)
        table.row['library'] = lib
        table.row['hash'] = hash_data(vec, lib, time, phi, n_fission_fragments)
        table.row['time'] = time
        table.row['phi'] = phi
        table.row['n_fission_fragments'] = n_fission_fragments
        table.row['execution time CRAM py_solve'] = CRAM_py_solve_time
        table.row['CRAM py_solve atom fraction'] = CRAM_py_solve_res
        CRAM_py_solve_res_normalized = CRAM_py_solve_res/np.sum(CRAM_py_solve_res)
        table.row['CRAM py_solve mass fraction'] = CRAM_py_solve_res_normalized
        table.row.append()
        table.flush()

def test_origen_against_CRAM_lambdify(xs_tape9, time, nuclide, phi):
    e_complex = CRAM_matrix_exp_lambdify()

    logger.info("Running CRAM lambdify %s at time=%s, nuclide=%s, phi=%s", xs_tape9, time, nuclide, phi)
    logger.info('-'*80)

    npzfilename = os.path.join('data', os.path.splitext(os.path.basename(xs_tape9))[0] + '_' +
    str(phi) + '.npz')

    nucs, mat = load_sparse_csr(npzfilename)
    assert mat.shape[1] == len(nucs)
    b = initial_vector(nuclide, nucs)

    CRAM_lambdify_time, CRAM_lambdify_res = time_func(e_complex, -mat*float(time), b)
    CRAM_lambdify_res = np.asarray(CRAM_lambdify_res)

    logger.info("CRAM lambdify runtime: %s", CRAM_lambdify_time)

    return CRAM_lambdify_time, CRAM_lambdify_res

def test_origen_against_CRAM_py_solve(xs_tape9, time, nuclide, phi):
    from py_solve.py_solve import expm_multiply14, asflat, N

    logger.info("Running CRAM pysolve %s at time=%s, nuclide=%s, phi=%s", xs_tape9, time, nuclide, phi)
    logger.info('-'*80)

    npzfilename = os.path.join('data', os.path.splitext(os.path.basename(xs_tape9))[0] + '_' +
    str(phi) + '.npz')

    nucs, mat = load_sparse_csr(npzfilename)
    assert mat.shape[1] == len(nucs) == N
    A = asflat(mat)
    b = initial_vector(nuclide, nucs)
    b = np.asarray(b, dtype='float64')

    CRAM_py_solve_time, CRAM_py_solve_res = time_func(expm_multiply14, -A*float(time), b)
    CRAM_py_solve_res = np.asarray(CRAM_py_solve_res)

    logger.info("CRAM py_solve runtime: %s", CRAM_py_solve_time)

    return CRAM_py_solve_time, CRAM_py_solve_res

def compute_mismatch(ORIGEN_data, CRAM_lambdify_res, CRAM_py_solve_res, nucs, rtol=1e-3, atol=1e-5):
    """
    Computes a mismatch analysis for an ORIGEN run vs. CRAM

    The default rtol is 1e-3 because ORIGEN returns 3 digits.

    The default atol is 1e-5 because ORIGEN stops the taylor expansion with
    the error term exp(ASUM)*ASUM**n/n! (using Sterling's approximation),
    where n = 3.5*ASUM + 6 and ASUM is the max of the column sums. The max of
    the column sums is ~2 because of fission, giving ~1e-5 (see ORIGEN lines
    5075-5100)

    """
    CRAM_lambdify_res_normalized = CRAM_lambdify_res/np.sum(CRAM_lambdify_res)
    CRAM_py_solve_res_normalized = CRAM_py_solve_res/np.sum(CRAM_py_solve_res)

    ORIGEN_res_weighted = origen_data_to_array_weighted(ORIGEN_data, nucs,)
    ORIGEN_res_materials = origen_data_to_array_materials(ORIGEN_data, nucs)
    # ORIGEN_res_atom_fraction = origen_data_to_array_atom_fraction(origen_data, nucs)

    for Cl, Cpy, O, units in [
        (CRAM_lambdify_res, CRAM_py_solve_res, ORIGEN_res_weighted, 'atom fractions'),
        (CRAM_lambdify_res_normalized, CRAM_py_solve_res_normalized, ORIGEN_res_materials, 'mass fractions'),
        # (CRAM_res_normalized, ORIGEN_res_atom_fraction, 'atom fraction'),
        ]:

        logger.info("Units: %s", units)
        d = {'CRAM lambdify': Cl, 'CRAM py_solve': Cpy, 'ORIGEN': O}
        for a_desc, b_desc in (['CRAM lambdify', 'CRAM py_solve'], ['CRAM lambdify', 'ORIGEN'], ['CRAM py_solve', 'ORIGEN']):
            a, b = d[a_desc], d[b_desc]
            try:
                np.testing.assert_allclose(a, b, rtol=rtol, atol=atol)
            except AssertionError as e:
                logger.info("%s and %s mismatch: %s", a_desc, b_desc, e)
                logger.info("Mismatching elements sorted by error (%s, %s, symmetric relative error):", a_desc, b_desc)
                D = np.isclose(a, b, rtol=rtol, atol=atol)
                rel_error = abs(a - b)/(a + b)
                for i, in np.argsort(rel_error, axis=0)[::-1]:
                    if D[i]:
                        continue
                    logger.info("%s %s %s %s", nucs[i], a[i], b[i], rel_error[i])
            else:
                logger.info("%s and %s arrays match with rtol=%s atol=%s", a_desc, b_desc, rtol, atol)

        logger.info('')

    # TODO: return some information here
def make_parser():
    p = argparse.ArgumentParser('origen', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('xs_tape9', metavar='xs-tape9', help="""path to the cross section TAPE9 file. If
    the path is not absolute, defaults to looking in {LIBS_DIR}""".format(LIBS_DIR=LIBS_DIR))
    p.add_argument('time', help='the time in sec',
                   type=float)
    p.add_argument('--phi', help='the neutron flux in [n/cm^2/sec]',
                   type=float, default=4e14)
    p.add_argument('--nuclide', help="The initial starting nuclide.",
        default="U235")
    p.add_argument('--decay-tape9', help="path to the decay TAPE9 file.",
        default=decay_TAPE9)
    p.add_argument('--origen', help="Path to the origen executable",
        default=ORIGEN)
    p.add_argument('--no-run-origen', action='store_false', dest='run_origen',
        help="Don't run origen")
    p.add_argument('--no-run-cram', action='store_false', dest='run_cram',
        help="Don't run cram")
    p.add_argument('--hdf5-file', default='data/results.hdf5', help="""hdf5 file
    to write results to""")
    return p

def execute(xs_tape9, time, phi, nuclide, hdf5_file='data/results.hdf5',
    decay_tape9=decay_TAPE9, origen=ORIGEN, run_origen=True,
    run_cram_lambdify=True, run_cram_py_solve=True):
    lib = os.path.splitext(os.path.basename(xs_tape9))[0]

    npzfilename = os.path.join('data', lib + '_' + str(phi) + '.npz')
    nucs, mat = load_sparse_csr(npzfilename)

    if run_origen:
        ORIGEN_time, ORIGEN_data = execute_origen(xs_tape9, time, nuclide, phi,
            origen, decay_tape9)
        test_origen_data_sanity(ORIGEN_data)
        save_file_origen(hdf5_file,
            ORIGEN_data=ORIGEN_data,
            lib=lib,
            nucs=nucs,
            start_nuclide=nuclide,
            time=time,
            phi=phi,
            ORIGEN_time=ORIGEN_time,
        )

    if run_cram_lambdify:
        CRAM_lambdify_time, CRAM_lambdify_res = test_origen_against_CRAM_lambdify(xs_tape9, time, nuclide, phi)
        save_file_cram_lambdify(hdf5_file,
            CRAM_lambdify_res=CRAM_lambdify_res,
            lib=lib,
            nucs=nucs,
            start_nuclide=nuclide,
            time=time,
            phi=phi,
            CRAM_lambdify_time=CRAM_lambdify_time,
        )

    if run_cram_py_solve:
        CRAM_py_solve_time, CRAM_py_solve_res = test_origen_against_CRAM_py_solve(xs_tape9, time, nuclide, phi)
        save_file_cram_py_solve(hdf5_file,
            CRAM_py_solve_res=CRAM_py_solve_res,
            lib=lib,
            nucs=nucs,
            start_nuclide=nuclide,
            time=time,
            phi=phi,
            CRAM_py_solve_time=CRAM_py_solve_time,
        )

    if run_origen and run_cram_lambdify and run_cram_py_solve:
        compute_mismatch(ORIGEN_data, CRAM_lambdify_res, CRAM_py_solve_res, nucs)

def main():
    p = make_parser()
    try:
        import argcomplete
        argcomplete.autocomplete(p)
    except ImportError:
        pass
    args = p.parse_args()
    execute(**vars(args))

if __name__ == '__main__':
    main()



import datetime
import difflib
import inspect
import os
import logging
import time
from functools import wraps
import io

import numpy as np
from scipy.sparse import csr_matrix
import mpmath
from sympy import lambdify, symbols

from sympy.utilities.decorator import conserve_mpmath_dps

t = symbols('t', real=True)

def diff_strs(a, b):
    """
    Print a colored character-by-character diff of a and b.

    Requires colorama to be installed.

    Differing characters from the a string are printed in red and those from
    the b string are printed in green.

    Note that this can be slow for large strings (more than 10000 characters).

    """
    try:
        import colorama
    except ImportError:
        raise ImportError("colorama is required to use diff_strs")

    s = difflib.SequenceMatcher(a=a, b=b, autojunk=False)
    for op, i1, j1, i2, j2 in s.get_opcodes():
        if op == 'equal':
            print(a[i1:j1], end='')
        elif op == 'replace':
            print(colorama.Fore.RED, a[i1:j1], colorama.Fore.GREEN, b[i2:j2], sep='', end='')
        elif op == 'insert':
            print(colorama.Fore.GREEN, b[i2:j2], sep='', end='')
        elif op == 'delete':
            print(colorama.Fore.RED, a[i1:j1], sep='', end='')
        print(colorama.Style.RESET_ALL, end='')
    print()

def relative_error(exact, approx):
    return abs(exact - approx)/exact

def mean_log10_relative_error(exact, approx):
    import numpy as np
    return np.mean(np.log10(abs(relative_error(exact, approx))))

# Function to create plot like in "Computing the Matrix Exponential in Burnup
# Calculations", Pusa and Leppänen:
# mpmath.cplot(lambdify(t, rat_func14 - exp(-t), 'mpmath'), re=[0, 100], im=[-30, 30], color=lambda i: -mpmath.floor(mpmath.log(abs(i), 10))/(30 - mpmath.floor(mpmath.log(abs(i), 10))), points=100000, verbose=True)

@conserve_mpmath_dps
def plot_in_terminal(expr, *args, prec=None, logname=None, **kwargs):
    """
    Run plot() but show in terminal if possible
    """
    from mpmath import plot
    if prec:
        mpmath.mp.dps = prec
    f = lambdify(t, expr, mpmath)
    try:
        from iterm2_tools.images import display_image_bytes
    except ImportError:
        if logname:
            os.makedirs('plots', exist_ok=True)
            file = 'plots/%s.png' % logname
        else:
            file = None
        plot(f, *args, file=file, **kwargs)
    else:
        from io import BytesIO
        b = BytesIO()
        plot(f, *args, **kwargs, file=b)
        if logname:
            os.makedirs('plots', exist_ok=True)
            with open('plots/%s.png' % logname, 'wb') as f:
                f.write(b.getvalue())
        print(display_image_bytes(b.getvalue()))

def plt_show_in_terminal(logname=None):
    import matplotlib.pyplot as plt
    try:
        from iterm2_tools.images import display_image_bytes
    except ImportError:
        plt.show()
    else:
        b = io.BytesIO()
        plt.savefig(b, format='png')
        print(display_image_bytes(b.getvalue()))
    if logname:
        os.makedirs('plots', exist_ok=True)
        plt.savefig('plots/%s.png' % logname, format='png')

def _get_log_file_name(locals_dict):
    d = locals_dict.copy()
    kwargs = d.pop('kwargs')
    d.update(kwargs)
    d.setdefault('maxsteps')
    d.setdefault('division')
    d.pop('log_to_file', None)
    degree = d.pop('degree')
    prec = d.pop('prec')
    info = 'degree=%s prec=%s ' % (degree, prec)
    info += ' '.join('%s=%s' % (i, d[i]) for i in sorted(d))
    return info


def log_function_args(func):
    """
    Decorator to log the arguments to func, and other info
    """
    logger = func.__globals__['logger']

    @wraps(func)
    def _func(*args, **kwargs):
        func_name = func.__name__
        logger.info("%s with arguments %s", func_name, args)
        logger.info("%s with keyword arguments %s", func_name, kwargs)

        os.makedirs('logs', exist_ok=True)
        binding = inspect.signature(func).bind(*args, **kwargs)
        binding.apply_defaults()
        logname = _get_log_file_name(binding.arguments)
        if kwargs.get('log_to_file', False):
            logger.addHandler(logging.FileHandler('logs/%s.log' % logname, delay=True))
        logger.info("Logging to file 'logs/%s.log'", logname)

        kwargs['logname'] = logname

        starttime = datetime.datetime.now()
        logger.info("Start time: %s", starttime)
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            logger.error("Exception raised", exc_info=True)
            raise
        finally:
            endtime = datetime.datetime.now()
            logger.info("End time: %s", endtime)
            logger.info("Total time: %s", endtime - starttime)

    return _func


def save_sparse_csr(filename, array, nucs, phi):
    """Saves a sparse CSR matrix to disk"""
    np.savez(filename, data=array.data, indices=array.indices,
             indptr=array.indptr, shape=array.shape, nucs=nucs,
             phi=phi)

def load_sparse_csr(filename):
    loader = np.load(filename)
    return (loader['nucs'], csr_matrix((loader['data'], loader['indices'], loader['indptr']),
                        shape=loader['shape']))


def time_func(f, *args, **kwargs):
    """
    Times f(*args, **kwargs)

    Returns (time_elapsed, f(*args, **kwargs)).

    """
    t = time.perf_counter()
    res = f(*args, **kwargs)
    return time.perf_counter() - t, res

def memoize(f):
    memo = {}
    @wraps(f)
    def inner(*args, **kwargs):
        hashable_kwargs = frozenset(kwargs.items())
        if (args, hashable_kwargs) not in memo:
            memo[args, hashable_kwargs] = f(*args, **kwargs)

        return memo[args, hashable_kwargs]

    return inner



from collections import defaultdict
import os
import argparse

import tables
import numpy as np
import matplotlib.pyplot as plt

from .tests.test_transmute import run_transmute_test
from .origen_all import TIME_STEPS
from .util import plt_show_in_terminal
from .cram import get_CRAM_from_cache, CRAM_coeffs

def analyze_origen(file):
    plt.clf()
    fig, ax = plt.subplots()

    times = {'ORIGEN': defaultdict(list), 'CRAM': defaultdict(list)}
    with tables.open_file(file, mode='r') as h5file:
        for run in 'ORIGEN', 'CRAM':
            for lib in h5file.root:
                table = h5file.get_node(lib, run.lower())
                for row in table:
                    times[run][row['time']].append(row['execution time ' + run])

            xvals = sorted(TIME_STEPS)

            x = []
            y = []
            for i, t in enumerate(xvals):
                itimes = times[run][sorted(times[run])[i]]
                x += [t]*len(itimes)
                y += itimes

            print("Longest", run, "runtime", max(y), "seconds")
            print("Shortest", run, "runtime", min(y), "seconds")

            ax.plot(x, y, 'o', label=run)

    # Tweak spacing to prevent clipping of tick-labels
    plt.subplots_adjust(bottom=0.15)
    plt.title("""\
Runtimes for ORIGEN and CRAM computing transmutation
over several starting libraries, nuclides, and timesteps.""")

    ax.set_xscale('log')
    ax.set_xticks(sorted(TIME_STEPS))
    ax.xaxis.set_ticklabels([TIME_STEPS[i].replace(' ', '\n') for i in
        sorted(TIME_STEPS)], size='small')
    ax.set_yscale('log')
    ax.legend()
    plt.ylabel('Runtime (seconds)')
    plt.xlabel('Time step t')

    plt_show_in_terminal()

def analyze_nofission():
    plt.clf()
    for time, time_name in sorted(TIME_STEPS.items()):
        nofission_transmutes = {}
        for f in os.listdir('data'):
            if f.endswith('_nofission.npz'):
                lib = f.split('_', 1)[0]
                data = os.path.join('data', f)
                print("analyzing", data, 'on', time_name)
                nofission_transmutes[lib] = run_transmute_test(data, 14, 200,
                    time, run_all=False, _print=True)

        for lib in nofission_transmutes:
            for r in nofission_transmutes[lib]:
                m = nofission_transmutes[lib][r]
                if not isinstance(m, np.ndarray):
                    m = m.toarray()
                if m is None or np.isnan(m).any() or np.isinf(m).any():
                    print("Could not compute", r, "for", lib)
                    continue
                title = lib + ' ' + r + ' ' + time_name
                plot_matrix_sum_histogram(m, title)

def plot_matrix_sum_histogram(m, title='', axis=0):
    plt.clf()
    plt.hist(np.asarray(np.sum(m, axis=axis)).flatten())
    plt.yscale('log', nonposy='clip')
    plt.title(title)
    plt_show_in_terminal()
    plt.close()

def analyze_cram_digits():
    print("Computing coefficients (or getting from cache)")
    exprs = defaultdict(dict)
    # {degree: {prec: {'p': [coeffs], 'q': [coeffs]}}}
    correct_digits = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    for degree in range(1, 21):
        print("Degree", degree)
        for prec in range(100, 1100, 100):
            print("Precision", prec)
            exprs[degree][prec] = CRAM_coeffs(get_CRAM_from_cache(degree,
                prec), prec)

        # Assume that 1000 has the most correct digits
        coeffs1000 = exprs[degree][1000]
        for prec in range(100, 1000, 100):
            coeffs = exprs[degree][prec]
            for l in 'pq':
                for coeff, coeff1000 in zip(coeffs[l], coeffs1000[l]):
                    correct_digits[degree][prec][l].append(len(os.path.commonprefix([coeff,
                        coeff1000])) - 1)


    # Plot minimum number of correct digits as a function of precision
    plt.clf()
    fig, ax = plt.subplots()

    minvals = defaultdict(list)
    for degree in range(1, 21):
        print("Degree", degree)
        for prec in range(100, 1000, 100):
            print("  Precision", prec)
            for l in 'pq':
                print('    ', end='')
                print(l, end=' ')
                for i in correct_digits[degree][prec][l]:
                    print(i, end=' ')
                print()
            minvals[degree].append(min(correct_digits[degree][prec]['p'] + correct_digits[degree][prec]['q']))
        ax.plot(range(100, 1000, 100), minvals[degree], label=degree)

    # TODO: Make window wider so the legend isn't chopped off
    ax.legend(title="Degree", loc="upper left", bbox_to_anchor=(1,1))
    plt.ylabel('Number of correct digits')
    plt.xlabel('Precision')

    plt_show_in_terminal()

    # Plot minimum number of correct digits as a function of degree
    plt.clf()
    fig, ax = plt.subplots()

    minvals = defaultdict(list)
    for prec in range(100, 1000, 100):
        for degree in range(1, 21):
            minvals[prec].append(min(correct_digits[degree][prec]['p'] + correct_digits[degree][prec]['q']))
        ax.plot(range(1, 21), minvals[prec], label=prec)

    # TODO: Make window wider so the legend isn't chopped off
    ax.legend(title="Precision", loc="upper left", bbox_to_anchor=(1,1))
    plt.ylabel('Number of correct digits')
    plt.xlabel('Degree')
    ax.set_xticks(range(1, 21))

    plt_show_in_terminal()


def analyze():
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--origen-results', default='data/results.hdf5',
        help="""HDF5 file for the results of the ORIGEN runs.""")
    parser.add_argument('--no-origen', action='store_false', dest='origen',
        help="""Don't run the origen analysis.""")
    parser.add_argument('--no-nofission', action='store_false',
        dest='nofission', help="""Don't run the nofission analysis.""")
    parser.add_argument('--cram-digits', action='store_true', help="""Analyze
        accuracy of CRAM digits. WARNING: If cache values have not been
        precomputed, this will take a long time (> 1 day) to compute.""")
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass
    args = parser.parse_args()

    if args.origen:
        analyze_origen(args.origen_results)
    if args.nofission:
        analyze_nofission()
    if args.cram_digits:
        analyze_cram_digits()

if __name__ == '__main__':
    analyze()



import warnings
import random

from sympy import (symbols, fraction, nsimplify, intervals, div, LC, Add,
    degree, re, im, together, expand_complex, Mul, I, nsolve, Function,
    Symbol, Poly, pi, count_roots)

from sympy.utilities.decorator import conserve_mpmath_dps

t = symbols('t', real=True)

@conserve_mpmath_dps
def thetas_alphas(rat_func, prec, *, use_intervals=False, eps=None):
    """
    Do a partial fraction decomposition of rat_func

    Returns (thetas, alphas, alpha0), where thetas and alphas are lists of
    values such that

        rat_func = alpha0 + sum([alpha/(t - theta) for theta,
            alpha in zip(thetas, alphas)])

    The thetas and alphas are in general complex numbers.

    Assumes that rat_func has the same degree numerator as denominator.

    If use_intervals=True, this uses the intevals() algorithm to do root
    finding. This algorithm is very slow, but has guaranteed precision, and is
    guaranteed to find all the roots. If it is False (the default), nsolve is
    used.

    eps is the length of the intervals for root finding. By default it is set
    to 10**-prec but it may need to be set smaller if there are roots smaller
    than ~1/10 to get full precision. If use_intervals=False, eps is ignored.
    """
    import mpmath
    mpmath.mp.dps = prec

    num, den = fraction(rat_func)
    d = degree(den)

    if use_intervals:
        rational_rat_func = nsimplify(rat_func)
        num, den = fraction(rational_rat_func)

        if d % 1:
            raise NotImplementedError("Odd degrees are not yet supported with use_intervals=True")

        # Note, eps is NOT the precision. It's the length of the interval.
        # If a root is small (say, on the order of 10**-N), then eps will need to be 10**(-N - d)
        # to get d digits of precision. For our exp(-t) approximations, the roots
        # (thetas) are all
        # within order 10**-1...10**1, so eps is *roughly* the precision.
        eps = eps or 10**-prec

        roots = intervals(den, all=True, eps=eps)[1]
        # eps ought to be small enough that either side of the interval is the
        # precision we want, but take the average (center of the rectangle)
        # anyway.
        # XXX: Make sure to change the evalf precision if eps is lowered.
        thetas = [((i + j)/2).evalf(prec) for ((i, j), _) in roots]
        # error = [(j - i).evalf(prec) for ((i, j), _) in roots]
    else:
        thetas = list(allroots(den, d, prec))
    alphas = []
    for theta in thetas:
        q, r = div(den, t - theta)
        alpha = (num/q).evalf(prec, subs={t: theta})
        alphas.append(alpha)
    alpha0 = (LC(num)/LC(den)).evalf(prec)
    return thetas, alphas, alpha0

def thetas_alphas_to_expr_real(thetas, alphas, alpha0):
    theta, alpha = symbols('theta, alpha')

    re_form = together(expand_complex(re(alpha/(t - theta))))

    return alpha0 + Add(*[re_form.subs({alpha: al, theta: th}) for th,
        al in zip(thetas, alphas)])

# sympy.re evaluates by default, which we don't want
class customre(Function):
    nargs = 1

    def _eval_evalf(self, prec):
        return re(self.args[0]._eval_evalf(prec))


def thetas_alphas_to_expr_complex(thetas, alphas, alpha0):
    return alpha0 + 2*customre(Add(*[alpha/(t - theta) for theta,
        alpha in zip(thetas, alphas) if im(theta) >= 0]))

def thetas_alphas_to_expr_complex2(thetas, alphas, alpha0):
    """
    Same as thetas_alphas_to_expr_complex except without the 2*re()
    optimization
    """
    return alpha0 + Add(*[alpha/(t - theta) for theta,
        alpha in zip(thetas, alphas)])

def allroots(expr, degree, prec, chop=True):
    roots = set()
    start = random.random() + random.random()*I
    MAX_ITERATIONS = 5*degree
    i = 0
    while len(roots) < degree:
        i += 1
        if i > MAX_ITERATIONS:
            raise RuntimeError("MAX_ITERATIONS exceeded. Could only find %s roots" % len(roots))
        try:
            r = nsolve(expr/Mul(*[t - r for r in roots]), start,
                maxsteps=1.7*prec, prec=prec)
        except ValueError:
            start = 10*(random.random() + random.random()*I)
        else:
            roots.add(r)

    # Because we started with a complex number, real roots will have a small
    # complex part. Assume that the roots with the smallest complex parts are
    # real. We could also check which roots are conjugate pairs.
    if chop:
        from sympy.core.compatibility import GROUND_TYPES
        if GROUND_TYPES == 'python':
            warnings.simplefilter('default')
            warnings.warn("It is recommended to install gmpy2 to speed up transmutagen")
        n_real_roots = count_roots(nsimplify(expr))
        roots = sorted(roots, key=lambda i:abs(im(i)))
        real_roots, complex_roots = roots[:n_real_roots], roots[n_real_roots:]
        real_roots = [re(i) for i in real_roots]
        roots = {*real_roots, *complex_roots}
    return roots

@conserve_mpmath_dps
def multiply_vector(expr, n0, horner=False):
    """
    Multiply the vector n0 into the expression expr

    It is recommended to make the vector noncommutative (Symbol("n0",
    commutative=False).

    This makes assumptions that the form of expr will be one of the ones
    generated by this file.

    To apply the horner scheme to the numerator and denominator, use horner=True.
    """
    # Make sure we are using a version of SymPy that has
    # https://github.com/sympy/sympy/pull/12088.
    x, y = symbols('x y')
    if Poly(pi.evalf(100)*x*y, x).as_expr() != pi.evalf(100)*x*y:
        raise RuntimeError("multiply_vector requires https://github.com/sympy/sympy/pull/12088")

    from sympy import horner as _horner
    if horner:
        # horner doesn't work on noncommutatives
        n1 = Symbol(n0.name)
        num, den = fraction(expr)
        return _horner(num*n1).subs(n1, n0)/_horner(den)

    # TODO: Don't distribute across complex numbers
    if expr.is_Add:
        if expr.is_number:
            return expr*n0
        # expand_mul(deep=False) does too much (it breaks horner)
        return Add(*[multiply_vector(i, n0) for i in expr.args])

    coeff, rest = expr.as_coeff_Mul()
    if isinstance(rest, customre):
        return coeff*customre(multiply_vector(rest.args[0], n0))

    num, den = fraction(expr)
    if expr != num:
        return multiply_vector(num, n0)/den
    return expr*n0



"""
Collect data for a matrix of ORIGEN and CRAM runs
"""
import argparse
import os
import logging
import datetime

from .tape9sparse import save_sparse
from .origen import execute, ORIGEN, decay_TAPE9, LIBS_DIR, logger as origen_logger

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
# Set to WARN for less output
logger.setLevel(logging.INFO)

ALL_LIBS = [
    # 'amo0ttta.lib',
    # 'amo0tttc.lib',
    'amo0tttr.lib',
    'amo1ttta.lib',
    'amo1tttc.lib',
    # 'amo1tttr.lib',
    'amo2ttta.lib',
    # 'amo2tttc.lib',
    'amo2tttr.lib',
    'amopttta.lib',
    'amoptttc.lib',
    'amoptttr.lib',
    # 'amopuuta.lib',
    'amopuutc.lib',
    # 'amopuutr.lib',
    'amopuuua.lib',
    'amopuuuc.lib',
    'amopuuur.lib',
    'amoruuua.lib',
    # 'amoruuuc.lib',
    'amoruuur.lib',
    # 'bwrpupu.lib',
    # 'bwrpuu.lib',
    'bwru.lib',
    # 'bwrue.lib',
    'bwrus.lib',
    # 'bwrus0.lib',
    # 'candunau.lib',
    'canduseu.lib',
    'crbra.lib',
    'crbrc.lib',
    # 'crbri.lib',
    'crbrr.lib',
    # 'emopuuua.lib',
    'emopuuuc.lib',
    # 'emopuuur.lib',
    'fftfc.lib',
    # 'pwrd5d33.lib',
    # 'pwrd5d35.lib',
    'pwrdu3th.lib',
    'pwrputh.lib',
    # 'pwrpuu.lib',
    'pwru.lib',
    'pwru50.lib',
    'pwrue.lib',
    'pwrus.lib'
]

INITIAL_NUCS = [
    'Th232',
    'U233',
    'U235',
    'U238',
    'Pu239',
    'Pu241',
    'Cm245',
    'Cf249',
    ]

DAY = 60*60*24
YEAR = 365.25*DAY
MONTH = YEAR/12
TIME_STEPS = {
    1:         '1 second',
    1*DAY:     '1 day',
    1*MONTH:   '1 month',
    1*YEAR:    '1 year',
    10*YEAR:   '10 years',
    1000*YEAR: '1000 years',
    1e6*YEAR:  '1 million years',
}

PHI = 4e14

def main():
    p = argparse.ArgumentParser('origen-all',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('--recompute-matrices', action='store_true',
        help="Recompute input matrices (npz files). Done automatically if they are not found.")
    p.add_argument('--decay-tape9', help="path to the decay TAPE9 file.",
        default=decay_TAPE9)
    p.add_argument('--origen', help="Path to the origen executable",
        default=ORIGEN)
    p.add_argument('--libs-dir', default=LIBS_DIR, help="Path to the libs/ directory")
    p.add_argument('--log-file', default='logs/origen_all.log', help='Path to file where output is logged')
    p.add_argument('--no-run-origen', action='store_false', dest='run_origen',
        help="Don't run origen")
    p.add_argument('--no-run-cram-lambdify', action='store_false', dest='run_cram_lambdify',
        help="Don't run cram lambdify")
    p.add_argument('--no-run-cram-py-solve', action='store_false', dest='run_cram_py_solve',
        help="Don't run cram py_solve")
    p.add_argument('--hdf5-file', default='data/results.hdf5', help="""hdf5 file
    to write results to""")

    args = p.parse_args()

    os.makedirs(os.path.dirname(args.log_file), exist_ok=True)
    logger.addHandler(logging.FileHandler(args.log_file, delay=True))
    origen_logger.addHandler(logging.FileHandler(args.log_file, delay=True))

    starttime = datetime.datetime.now()
    logger.info("Start time: %s", starttime)

    try:
        npzfilenames = []
        for tape9 in ALL_LIBS:
            logger.info("Computing library %s", tape9)
            xs_tape9 = os.path.join(args.libs_dir, tape9)

            base = os.path.basename(xs_tape9)
            base, _ = os.path.splitext(base)
            os.makedirs('data', exist_ok=True)
            npzfilename = os.path.join('data', base + '_' + str(PHI) + '.npz')
            npzfilenames.append(npzfilename)

        if args.recompute_matrices or not any(os.path.exists(npzfilename) for
            npzfilename in npzfilenames):
            logger.info("Saving matrices")
            save_sparse(args.libs_dir, phi=PHI, output=npzfilename,
                decaylib=args.decay_tape9)

        for tape9, npzfilename in zip(ALL_LIBS, npzfilenames):
            logger.info("Computing library %s", tape9)
            xs_tape9 = os.path.join(args.libs_dir, tape9)

            for initial_nuclide in INITIAL_NUCS:
                logger.info("Using initial nuclide %s", initial_nuclide)
                for time in sorted(TIME_STEPS):
                    logger.info("Using time %s (%s s)", TIME_STEPS[time], time)
                    logger.info("Run: %s %s %s", tape9, initial_nuclide, time)
                    try:
                        execute(xs_tape9, time, PHI, initial_nuclide,
                            decay_tape9=args.decay_tape9, origen=args.origen,
                            run_cram_lambdify=args.run_cram_lambdify,
                            run_cram_py_solve=args.run_cram_py_solve,
                            run_origen=args.run_origen,
                            hdf5_file=args.hdf5_file)
                    except AssertionError as e:
                        logger.error("AssertionError with lib %s: %s", tape9, e)
    except BaseException as e:
        logger.error("Exception raised", exc_info=True)
        raise
    finally:
        endtime = datetime.datetime.now()
        logger.info("End time: %s", endtime)
        logger.info("Total time: %s", endtime - starttime)


if __name__ == '__main__':
    main()



"""Converts a TAPE9 file to a sparse matrix on disk."""
import os
import argparse

from .util import save_sparse_csr
from .tape9utils import tape9_to_sparse, THRESHOLD, normalize_tape9s

def make_parser():
    p = argparse.ArgumentParser('tape9sparse', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('tape9s', nargs='+', help="""Paths to the TAPE9 files. If a
    path is a directory, a set of default libraries will be gathered from that
    directory (transmutagen.origen_all.ALL_LIBS)""")
    p.add_argument('--phi', help='the neutron flux in [n/cm^2/sec]',
                   type=float, default=4e14)
    p.add_argument('-f', '--format', help='The sparse matrix format',
                   default='csr', dest='format')
    p.add_argument('-d', '--decay', help='path to the decay file, if needed',
                   default='decay.lib', dest='decaylib')
    p.add_argument('--include-fission', action='store_true', default=True,
                   dest='include_fission',
                   help='Include fission reactions in the matrix.')
    p.add_argument('--no-include-fission', action='store_false',
                   dest='include_fission',
                   help="Don't include fission reactions in the matrix.")
    p.add_argument('-t', '--threshold', default=THRESHOLD, dest='threshold',
                   help='cutoff for ignoring reactions', type=float)
    p.add_argument('-o', '--output-dir', default=None,
                   help='The directory to write the output files to, in npz format.')
    return p

def save_sparse(tape9s, phi=4e14, output_dir=None, format='csr',
    decaylib='decay.lib', include_fission=True, threshold=THRESHOLD):
    if output_dir is None:
        output_dir = 'data'
    os.makedirs(output_dir, exist_ok=True)

    if format != 'csr':
        raise ValueError('Only the CSR format is currently available from the '
                         'command line interface.')

    tape9s = normalize_tape9s(tape9s)
    mats, nucs = tape9_to_sparse(tape9s, phi, format=format,
                                decaylib=decaylib,
                                include_fission=include_fission,
                                threshold=threshold)
    for tape9, mat in zip(tape9s, mats):
        base = os.path.basename(tape9)
        base, _ = os.path.splitext(base)
        fission_part = '' if include_fission else '_nofission'
        output = os.path.join('data', base + '_' + str(phi) + fission_part + '.npz')

        print("Writing file to", output)
        save_sparse_csr(output, mat, nucs, phi)

def main(args=None):
    p = make_parser()
    ns = p.parse_args(args=args)
    save_sparse(**vars(ns))

if __name__ == '__main__':
    main()



# PYTHON_ARGCOMPLETE_OK
import argparse
import os
import logging
import ast
import random

import mpmath
from sympy import (nsolve, symbols, Mul, Add, chebyshevt, exp, simplify,
    chebyshevt_root, Tuple, diff, N, solve, Poly, lambdify, sign, fraction,
    sympify, Float, srepr, Rational, log)

from sympy.utilities.decorator import conserve_mpmath_dps

from .util import plot_in_terminal, plt_show_in_terminal, log_function_args

# Give a better error message if not using SymPy master
try:
    Poly.transform
except AttributeError:
    raise ImportError("transmutagen requires the git master version of SymPy")

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
# Change to INFO or DEBUG for more output
logger.setLevel(logging.WARN)

t = symbols('t', real=True)

def general_rat_func(degree, x, chebyshev=False):
    """
    Return a general rational function with numerator and denominator degree d

    Returns a tuple, (rational function, numerator coefficients, denominator coefficients)

    The constant coefficient in the denominator is always 1.
    """
    if isinstance(degree, (list, tuple)):
        num_degree, den_degree = degree
    else:
        num_degree = den_degree = degree

    num_coeffs = symbols('p:%s' % (num_degree+1))
    den_coeffs = symbols('q:%s' % (den_degree+1))[1:]
    if chebyshev:
        num = Add(*(Mul(c, chebyshevt(i, x)) for i, c in enumerate(num_coeffs)))
        den = Add(*(Mul(c, chebyshevt(i, x)) for i, c in enumerate([1, *den_coeffs])))
        rat_func = num/den
    else:
        rat_func = Poly(reversed(num_coeffs), x)/Poly([*reversed(den_coeffs), 1], x)
    return rat_func, num_coeffs, den_coeffs

@conserve_mpmath_dps
def nsolve_intervals(expr, bounds, division=10000, solver='bisect', scale=True, prec=None, **kwargs):
    """
    Divide bounds into division intervals and nsolve in each one
    """
    roots = []
    L = bounds[1] - bounds[0]
    # These are only needed for scaling and sign checks, so don't bother with
    # full precision
    points = [bounds[0] + i*L/division for i in range(division+1)]

    expr_values = []
    mpmath.mp.dps = prec
    f = lambdify(t, expr, 'mpmath')
    for i, point in enumerate(points):
        if not i % 10000:
            logger.debug("expr_values %s/%s", i, division)
        expr_values.append(f(mpmath.mpf(point)))

    for i in range(division):
        interval = [bounds[0] + i*L/division, bounds[0] + (i + 1)*L/division]
        try:
            logger.debug("Solving in interval %s", interval)
            s1 = expr_values[i]
            s2 = expr_values[i+1]
            if sign(s1) == sign(s2):
                logger.debug("Expr doesn't change signs on %s, skipping", interval)
                # logger.debug("Expr values, %s, %s", s1, s2)
                continue

            if scale:
                val = expr_values[i]
                logger.debug("Scaling by %s", val)
                scaled_expr = expr/val
            else:
                scaled_expr = expr

            root = nsolve(scaled_expr, interval, solver=solver, prec=prec, **kwargs)
        except ValueError as e:
            logger.debug("No solution found: %s", e)
            continue
        else:
            if interval[0] < root < interval[1]:
                logger.debug("Solution found: %s", root)
                roots.append(root)
                if sign(s1) == sign(s2):
                    logger.debug("Root found even though signs did not change")
            else:
                logger.warn("%s is not in %s, discarding", root, interval)

    return roots

def nsolve_points(expr, bounds, division=300, scale=True, **kwargs):
    """
    Divide bounds into division points and nsolve near each one
    """
    roots = []
    L = bounds[1] - bounds[0]
    for i in range(division):
        point = bounds[0] + i*L/division
        try:
            logger.debug("Solving near point %s", point)
            if scale:
                val = expr.evalf(kwargs['prec'], subs={t:point})
                logger.debug("Scaling by %s", val)
                expr /= val
            root = nsolve(expr, point, **kwargs)
        except ValueError as e:
            logger.debug("No solution found: %s", e)
            continue
        else:
            if root not in roots:
                logger.debug("Solution found: %s", root)
                roots.append(root)
            else:
                logger.debug("%s already found, discarding", root)

    return sorted(roots)

@conserve_mpmath_dps
@log_function_args
def CRAM_exp(degree, prec=128, *, max_loops=30, c=None, maxsteps=None,
    convergence_value=None, tol=None, nsolve_type='intervals', D_scale=1,
    plot=False, log_to_file=False, seed=None, initial_points=None, **kwargs):
    """
    Compute the CRAM approximation of exp(-t) from t in [0, oo) of the given degree

    The Remez algorithm is used.

    degree is the degree of the numerator and denominator of the
    approximation, or a list of [numerator degree, denominator degree]. Note
    that list degrees are still not fully supported.

    prec is the precision of the floats used in the calculation. Note that, as
    of now, the result may not be accurate to prec digits.

    max_loops is the number of loops to run the Remez algorithm before giving
    up (default 20).

    c is the coefficient of the transformation c*(t + 1)/(t - 1) that is
    applied to exp(-t) to translate it to the interval [-1, 1]. It should be
    chosen so that the maximal errors of the approximation are equally spaced
    in that interval. The default is 0.6*degree.

    maxsteps is the argument passed to the mpmath bisection solver. The
    default is 1.7*prec. See also
    https://github.com/fredrik-johansson/mpmath/issues/339.

    convergence_value is the value where the algorithm converges if the
    maximum absolute error minus the minimum absolute error is below this
    value. The default is to converge automatically (when the max - min values
    are within 10 orders of magnitude of the precision, and become
    log-convex).

    tol is the tolerance passed to nsolve. The default is 10**-(prec - 7).

    nsolve_type can be 'points' or 'intervals'.

    D_scale is a factor used to scale the derivative before root finding.

    initial_points should be 'chebyshev', in which case the initial points are
    the chebyshev nodes, 'random', in which case the points are random, or
    None (the default), which chooses automatically ('chebyshev' for order <
    28, 'random' for order >= 28).

    seed is the random seed for initial_points='random'. The default is to
    pick a random seed.

    Additional keyword arguments are passed to nsolve_intervals, such as
    division and scale.

    If plot=True, plots are shown of the iterations. If iterm2_tools is
    installed, plots are shown inline in the terminal.

    If log_to_file=True, logging output is also logged to a file (in logs/), and
    plots are also saved in plots/.

    The SymPy master branch is required for this to work.

    """
    # From the log_function_args decorator
    logname = kwargs['logname']

    if not seed:
        seed = random.randint(0, 2**64)

    logger.info("Random seed: %s", seed)

    R = random.Random(seed)

    epsilon, i, y = symbols("epsilon i y")

    if isinstance(degree, (list, tuple)):
        num_degree, den_degree = degree
    else:
        num_degree = den_degree = degree

    c = c or 0.6*max(num_degree, den_degree)

    maxsteps = int(maxsteps or 1.7*prec)
    tol = tol or mpmath.mpf(10)**-(prec - 8)

    if nsolve_type == 'points':
        nsolve_func = nsolve_points
    elif nsolve_type == 'intervals':
        nsolve_func = nsolve_intervals
    else:
        raise ValueError("nsolve_type must be 'points' or 'intervals'")

    r, num_coeffs, den_coeffs = general_rat_func(degree, t, chebyshev=True)
    E = exp(c*(t + 1)/(t - 1)) - r
    expr = E + (-1)**i*epsilon
    expr = expr*r.as_numer_denom()[1]
    expr = simplify(expr)

    if initial_points is None:
        initial_points = 'chebyshev' if max(num_degree, den_degree) < 28 else 'random'
    if initial_points == 'chebyshev':
        points = [chebyshevt_root((num_degree + 1) + (den_degree + 1) + 1,
            (num_degree + 1) + (den_degree + 1) - j) for j in
            range(1, (num_degree + 1) + (den_degree + 1) + 1)]
    elif initial_points == 'random':
        points = [Float(R.uniform(-1, 1)) for i in range(1, (num_degree + 1) + (den_degree + 1) + 1)]
        points.sort(key=Rational)
    else:
        raise ValueError("initial_points must be 'chebyshev' or 'random'")
    points = [i.evalf(prec) for i in points]
    maxmins = []
    for iteration in range(max_loops):
        logger.info('-'*80)
        logger.info("Iteration %s:", iteration)
        system = Tuple(*[expr.subs({i: j, t: points[j]}) for j in range(num_degree + den_degree + 1)])
        system = system + Tuple(expr.replace(exp, lambda i: 0).subs({i: num_degree + den_degree + 1, t: 1}))
        logger.debug('system: %s', system)
        logger.debug('[*num_coeffs, *den_coeffs, epsilon]: %s', [*num_coeffs, *den_coeffs, epsilon])
        sol = dict(zip([*num_coeffs, *den_coeffs, epsilon], nsolve(system,
            [*num_coeffs, *den_coeffs, epsilon], [*[1]*((num_degree + 1) + (den_degree + 1) - 1), 0],
            prec=prec, maxsteps=maxsteps)))
        logger.info('sol: %s', sol)
        logger.info('system.subs(sol): %s', [i.evalf() for i in system.subs(sol)])
        D = diff(E.subs(sol), t)
        if plot:
            plot_in_terminal(E.subs(sol), (-1, 0.999), prec=prec, points=1000,
                logname=log_to_file and logname + ' iteration=%s' % iteration)
        logger.info('E.subs(sol): %s', E.subs(sol))

        D *= D_scale
        # we can't use 1 because of the singularity
        points = [-1, *nsolve_func(D, [-1, 0.999999], prec=prec, tol=tol, maxsteps=maxsteps, **kwargs), 1]
        logger.debug('points: %s', points)
        logger.info('D: %s', D)
        logger.info('[(i, D.subs(t, i)) for i in points]: %s', [(i, D.subs(t, i)) for i in points])
        if not len(points) == (num_degree + 1) + (den_degree + 1):
            raise RuntimeError("ERROR: len(points) is (%s), not 2*(degree + 1) (%s)" % (len(points), (num_degree + 1) + (den_degree + 1)))
        Evals = [E.evalf(prec, subs={**sol, t: point}) for point in points[:-1]] + [-r.evalf(prec, subs={**sol, t: 1})]
        logger.info('Evals: %s', Evals)
        # Rational is needed because of a SymPy issue with sorting Floats. See
        # https://github.com/sympy/sympy/issues/11707#issuecomment-292623480.
        absEvals = list(map(abs, map(Rational, Evals)))
        maxmin = N(max(absEvals) - min(absEvals))
        maxmins.append(maxmin)
        logger.info('max - min: %s', maxmin)
        logger.info('epsilon: %s', N(sol[epsilon]))
        if plot:
            import matplotlib.pyplot as plt
            plt.clf()
            fig, ax = plt.subplots()
            ax.plot(range(iteration+1), maxmins, linestyle='-', marker='o')
            ax.set_yscale('log')
            ax.set_xticks(range(iteration+1))
            plt.xlabel("Iteration")
            plt.ylabel("max - min")
            plt.title("Convergence")
            plt_show_in_terminal(logname=log_to_file and logname + ' iteration=%s convergence' % iteration)

        converged = False
        if convergence_value:
            if maxmin < convergence_value:
                converged = True
        else:
            # Once max - min is sufficiently close to the precision, converge
            # when the max - min values become log-convex.
            if (iteration >= 2 and maxmin < mpmath.mpf(10)**-(prec - 10) and
               log(maxmins[-2]) - log(maxmins[-3]) < log(maxmins[-1]) - log(maxmins[-2])):
                converged = True
        if converged:
            logger.info("Converged in %d iterations.", iteration + 1)
            break
    else:
        logger.warn("!!!WARNING: DID NOT CONVERGE AFTER %d ITERATIONS!!!", max_loops)


    # Workaround an issue. Poly loses precision unless the mpmath precision is
    # set.
    mpmath.mp.dps = prec

    # We need to be very careful about doing the inverse Mobius
    # transformation. Poly.transform does the transformation without losing
    # precision. See SymPy issue https://github.com/sympy/sympy/issues/12003.
    n, d = map(Poly, fraction(r.subs(sol)))
    inv = solve(-c*(t + 1)/(t - 1) - y, t, rational=False)[0].subs(y, t)
    p, q = map(lambda i: Poly(i, t), fraction(inv))
    n, d = n.transform(p, q), d.transform(p, q)
    if num_degree > den_degree:
        d = d*q**(num_degree - den_degree)
    elif den_degree > num_degree:
        n = n*q**(den_degree - num_degree)
    rat_func = n/d.TC()/(d/d.TC())
    ret = rat_func.evalf(prec)

    logger.info('rat_func: %s', rat_func)
    if plot:
        plot_in_terminal(rat_func - exp(-t), (0, 100), prec=prec, points=1000,
            logname=log_to_file and logname + ' final')

    return ret

def get_CRAM_from_cache(degree, prec, expr=None, plot=False, log=False, use_cache=True):
    if log:
        if log is True:
            log = "INFO"
        logger.setLevel(getattr(logging, log.upper()))
        log_to_file = True
    else:
        log_to_file = False

    if use_cache:
        installed_cache = os.path.join(os.path.dirname(__file__), 'data', 'CRAM_cache')
        local_cache = os.path.expanduser('~/.transmutagen/CRAM_cache')
        os.makedirs(local_cache, exist_ok=True)
        installed_cache_file = os.path.join(installed_cache, '%s_%s' % (degree, prec))
        local_cache_file = os.path.join(local_cache, '%s_%s' % (degree, prec))

    if use_cache and not expr:
        if os.path.exists(installed_cache_file):
            with open(installed_cache_file) as f:
                expr = sympify(f.read(), globals())
        elif os.path.exists(local_cache_file):
            with open(local_cache_file) as f:
                expr = sympify(f.read(), globals())
        else:
            expr = None

    expr = expr or CRAM_exp(degree, prec, plot=plot, log_to_file=log_to_file)
    if use_cache:
        with open(local_cache_file, 'w') as f:
            f.write(srepr(expr))

    return expr

def CRAM_coeffs(expr, prec, decimal_rounding=False, truncate=False):
    """
    Returns a dictionary of the coefficients from expr, as strings

    The dictionary keys are 'p' and 'q'. 'p' has a list of the numerator
    coefficients and 'q' has a list of the denominator coefficients. The
    coefficients are listed from 0th (constant) to nth.

    The coefficients are rounded to prec, regardless of their original
    precision.

    If decimal_rounding is True (the default is False), rounding to prec will
    be done using decimal arithmetic instead of binary arithmetic.

    If truncate is True (the default is False), the value is truncated instead
    of rounded (decimal_rounding is ignored in this case).

    """
    import decimal
    coeffs = {}
    n, d = fraction(expr)
    _p = list(reversed(Poly(n).coeffs()))
    _q = list(reversed(Poly(d).coeffs()))
    p = []
    q = []
    assert len(_p) == len(_q)

    format_str = '{:.%se}' % (prec - 1)
    for _l, l in zip([_p, _q], [p, q]):
        for i in _l:
            if truncate:
                i_prec = mpmath.libmp.prec_to_dps(i._prec)
                s = ('{:.%se}'%i_prec).format(i)
                m, e = s.split('e')
                length = prec + 1
                if '-' in m:
                    length += 1
                l.append(m[:length] + 'e' + e)
            elif decimal_rounding:
                l.append(format_str.format(decimal.Decimal(repr(i))))
            else:
                l.append(format_str.format((Float(i, prec))))

    coeffs['p'] = p
    coeffs['q'] = q
    return coeffs

def main():
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('degree')
    parser.add_argument('prec', type=int)
    parser.add_argument('--division', type=int)
    parser.add_argument('--c', type=float)
    parser.add_argument('--maxsteps', type=int)
    parser.add_argument('--max-loops', type=int)
    parser.add_argument('--convergence-value', type=mpmath.mpf)
    parser.add_argument('--tol', type=mpmath.mpf)
    parser.add_argument('--nsolve-type', default=None,
        choices=['points', 'intervals'])
    parser.add_argument('--solver', default=None)
    parser.add_argument('--D-scale', default=None, type=float)
    parser.add_argument('--scale', default=None, type=bool)
    parser.add_argument('--plot', default=True, type=bool)
    parser.add_argument('--seed', default=None, type=int)
    parser.add_argument('--initial-points', default=None,
        choices=['chebyshev', 'random'])
    parser.add_argument('--log-to-file', default=True, type=bool,
        help="Log output to a file (in the logs/ directory)")
    parser.add_argument('--log-level', default='info',
        choices=['debug', 'info', 'warning', 'error', 'critical'])
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass
    args = parser.parse_args()

    try:
        args.degree = ast.literal_eval(args.degree)
    except (ValueError, SyntaxError) as e:
        parser.error("Invalid degree: " + str(e))

    arguments = args.__dict__.copy()
    for i in arguments.copy():
        if not arguments[i]:
           del arguments[i]
    if args.log_level:
        logger.setLevel(getattr(logging, args.log_level.upper()))
        del arguments['log_level']

    CRAM_exp(**arguments)

if __name__ == '__main__':
    main()



from functools import wraps

from sympy import Mul, symbols, lambdify, Add, LC, fraction

from sympy.printing.lambdarepr import NumPyPrinter
from sympy.printing.precedence import precedence

import numpy as np
import scipy.sparse.linalg

from .cram import CRAM_exp, get_CRAM_from_cache
from .partialfrac import (thetas_alphas, thetas_alphas_to_expr_complex,
    thetas_alphas_to_expr_real, t, multiply_vector, allroots)
from .util import memoize

class MatrixNumPyPrinter(NumPyPrinter):
    """
    Print an expression for numpy assuming the variables are matrices

    Prints inversions as solve() and multiplication of nonconstants as @.

    """
    _default_settings = {
        **NumPyPrinter._default_settings,
        # TODO: Make this automatic
        'use_autoeye': True,
        'py_solve': False,
        }

    def __init__(self, settings=None):
        if settings is not None and 'py_solve' in settings:
            if settings.get('use_autoeye', False):
                raise ValueError("use_autoeye cannot be used with py_solve")
            settings['use_autoeye'] = False
        super().__init__(settings)

    def _print_Add(self, expr):
        if not (self._settings['use_autoeye'] or self._settings['py_solve']):
            return super()._print_Add(expr)

        prec = precedence(expr)

        num_terms = [i for i in expr.args if i.is_number]
        rest_terms = [i for i in expr.args if i not in num_terms]

        if len(rest_terms) > 1:
            rest = super()._print_Add(Add(*rest_terms))
        elif len(rest_terms) == 1:
            rest = self._print(rest_terms[0])
        else:
            if self._settings['py_solve']:
                return super()._print_Add(expr)
            rest = ''

        if len(num_terms) > 1:
            num = self.__class__({**self._settings, 'use_autoeye': False})._print_Add(Add(*num_terms))
        elif len(num_terms) == 1:
            num = self.__class__({**self._settings, 'use_autoeye': False})._print(num_terms[0])
        else:
            num = ''

        if rest and num:
            if self._settings['py_solve']:
                return "diag_add(%s, %s)" % (self._print(rest_terms[0]), self._print(Add(*num_terms)))
            return self.parenthesize(rest + ' + autoeye(%s)' % num, prec)
        elif rest:
            return self.parenthesize(rest, prec)
        else:
            if self._settings['use_autoeye']:
                # No need to parenthesize
                return 'autoeye(%s)' % num
            else:
                return self.parenthesize(num, prec)

    def _print_Mul(self, expr):
        prec = precedence(expr)

        pows = [i for i in expr.args if i.is_Pow and i.exp < 0]
        if len(pows) > 1:
            raise NotImplementedError("Need exactly one inverted Pow, not %s" % len(pows))

        if not pows:
            no_autoeye = self.__class__({**self._settings, 'use_autoeye': False})
            num_terms = [no_autoeye._print(no_autoeye.parenthesize(i, prec)) for i in
                expr.args if i.is_number]
            mat_terms = [self._print(self.parenthesize(i, prec)) for i in
                expr.args if not i.is_number]
            if len(mat_terms) >= 2 and self._settings['py_solve']:
                raise NotImplementedError("matrix multiplication is not yet supported with py_solve")
            if num_terms and mat_terms:
                return '*'.join(num_terms) + '*' + '@'.join(mat_terms)
            else:
                if self._settings['use_autoeye']:
                    if num_terms:
                        return ('autoeye(%s)' % '*'.join(num_terms)) + '@'.join(mat_terms)
                    return '@'.join(mat_terms)

                return '*'.join(num_terms) + '@'.join(mat_terms)

        [pow] = pows

        rest = Mul(*[i for i in expr.args if i != pow])

        return 'solve(%s, %s)' % (self._print(1/pow), self._print(rest))

    def _print_Integer(self, expr):
        if self._settings['use_autoeye']:
            return 'autoeye(%s)' % super()._print_Integer(expr)
        return super()._print_Integer(expr)

    def _print_Float(self, expr):
        if self._settings['use_autoeye']:
            return 'autoeye(%s)' % super()._print_Float(expr)
        return super()._print_Float(expr)

    def _print_Pow(self, expr):
        if self._settings['py_solve']:
            raise NotImplementedError("Matrix powers are not yet supported with py_solve")
        if expr.exp.is_Integer and expr.exp > 1:
            return 'matrix_power(%s, %s)' % (self._print(expr.base), expr.exp)
        return super()._print_Pow(expr)

    def _print_ImaginaryUnit(self, expr):
        if self._settings['use_autoeye']:
            return 'autoeye(1j)'
        return '1j'

    def _print_customre(self, expr):
        return 'real(%s)' % self._print(expr.args[0])

class autoeye:
    __array_priority__ = 11

    def __init__(self, coeff=1):
        self.coeff = coeff

    def eval(self, shape, eye_type, dtype=None):
        return self.coeff*eye_type(shape, dtype=dtype)

    def __eq__(self, other):
        if isinstance(other, autoeye):
            return self.coeff == other.coeff
        return False

    def __add__(self, other):
        if isinstance(other, autoeye):
            return autoeye(self.coeff + other.coeff)

        if isinstance(other, (int, float, complex)):
            return autoeye(self.coeff + other)

        if isinstance(other, np.ndarray):
            eye_type = np.eye
        elif isinstance(other, scipy.sparse.spmatrix):
            eye_type = scipy.sparse.eye
        else:
            return NotImplemented

        if len(other.shape) != 2:
            raise ValueError("autoeye can only be added to 2-dim numpy arrays")

        if other.shape[0] != other.shape[1]:
            raise ValueError("autoeye can only be added to square numpy arrays, other.shape is %s" % (other.shape,))

        return self.eval(other.shape[0], eye_type, dtype=other.dtype) + other

    __radd__ = __add__

    def __mul__(self, other):
        if isinstance(other, autoeye):
            return autoeye(self.coeff * other.coeff)

        if isinstance(other, (int, float, complex)):
            return autoeye(self.coeff * other)

        return NotImplemented

    __rmul__ = __mul__

    def __matmul__(self, other):
        if isinstance(other, autoeye):
            return autoeye(self.coeff * other.coeff)

        if isinstance(other, np.ndarray):
            eye_type = np.eye
        elif isinstance(other, scipy.sparse.spmatrix):
            eye_type = scipy.sparse.eye
        else:
            return NotImplemented

        if len(other.shape) != 2:
            raise ValueError("autoeye can only be matmuled by 2-dim numpy arrays")

        return self.eval(other.shape[0], eye_type=eye_type, dtype=other.dtype) @ other

    def __rmatmul__(self, other):
        if isinstance(other, autoeye):
            return autoeye(self.coeff * other.coeff)

        if isinstance(other, np.ndarray):
            eye_type = np.eye
        elif isinstance(other, scipy.sparse.spmatrix):
            eye_type = scipy.sparse.eye
        else:
            return NotImplemented

        if len(other.shape) != 2:
            # Matmul works weird on 1d arrays. It treats them as column
            # vectors from the left and row vectors from the right.
            raise ValueError("autoeye can only be matmuled by 2-dim numpy arrays")

        ret_shape = other.shape[1]
        return other @ self.eval(ret_shape, eye_type=eye_type, dtype=other.dtype)

    def __str__(self):
        return 'autoeye(%s)' % self.coeff

    __repr__ = __str__

def numpy_solve_with_autoeye(a, b, **kwargs):
    if isinstance(a, autoeye):
        a = a.eval(b.shape[0], np.eye)
    if isinstance(b, autoeye):
        b = b.eval(a.shape[0], np.eye)

    return np.linalg.solve(a, b, **kwargs)


def scipy_sparse_solve_with_autoeye(a, b, **kwargs):
    if isinstance(a, autoeye):
        a = a.eval(b.shape[0], scipy.sparse.eye)
    if isinstance(b, autoeye):
        b = b.eval(a.shape[0], scipy.sparse.eye)

    ret = scipy.sparse.linalg.spsolve(a, b, **kwargs)
    if isinstance(ret, np.ndarray):
        ret = ret[:,np.newaxis]

    return ret

scipy_translations = {
    'solve': scipy.sparse.linalg.spsolve,
    'autoeye': autoeye,
    'matrix_power': lambda a, b: a**b,
    'real': lambda m: np.real(m) if isinstance(m, np.ndarray) else scipy.sparse.csr_matrix((np.real(m.data), m.indices,
        m.indptr), shape=m.shape),
    }

scipy_translations_autoeye = {
    **scipy_translations,
    'solve': scipy_sparse_solve_with_autoeye,
    }

@memoize
def CRAM_matrix_exp_lambdify(degree=14, prec=200, use_cache=True,
    form='complex partial fraction', py_solve=False):
    """
    Return a lambdified function for the CRAM approximation to exp(-x)

    form can be one of

    'complex partial fraction' (the default)
    'real partial fraction'
    'rational function'
    'rational function horner'
    'factored'

    When py_solve = True, the py_solve module will be used (scipy is used
    otherwise). In this case, it is much faster to pre-flatten the input
    matrix:

    >>> mat, time, b = ...
    >>> mat = py_solve.asflat(mat)
    >>> f = CRAM_matrix_exp_lambdify(py_solve=True)
    >>> f(-mat*time, b)

    """
    # TODO: This function should give exp(x), not exp(-x)

    if use_cache:
        rat_func = get_CRAM_from_cache(degree, prec)
    else:
        rat_func = CRAM_exp(degree, prec, plot=False)
    thetas, alphas, alpha0 = thetas_alphas(rat_func, prec)
    if form == 'complex partial fraction':
        expr = thetas_alphas_to_expr_complex(thetas, alphas, alpha0)
    elif form == 'real partial fraction':
        expr = thetas_alphas_to_expr_real(thetas, alphas, alpha0)
    elif form in ['rational function', 'rational function horner']:
        expr = rat_func
    elif form == 'factored':
        num, den = fraction(rat_func)
        # XXX: complex conjugate roots have the same absolute value
        numroots = sorted(allroots(num, degree, prec), key=lambda i: abs(i))
        denroots = sorted(allroots(den, degree, prec), key=lambda i: abs(i))
        p1q1 = LC(num)/LC(den)
    else:
        raise ValueError("Invalid argument for 'form': %s" % (form,))
    n0 = symbols("n0", commutative=False)

    if py_solve:
        from py_solve import py_solve
        module = [py_solve, 'numpy']
        printer = MatrixNumPyPrinter({'py_solve': True})
        def wrapper(f):
            @wraps(f)
            def _f(t, n0):
                t = py_solve.asflat(t)
                return f(t, n0)
            return _f
    else:
        module = scipy_translations_autoeye
        printer = MatrixNumPyPrinter({'use_autoeye': True})
        wrapper = lambda f: f

    if form != 'factored':
        return wrapper(lambdify((t, n0), multiply_vector(expr, n0,
            horner=(form == 'rational function horner')),
            module, printer=printer, dummify=False))
    else:
        if py_solve:
            raise NotImplementedError("py_solve is not supported with factor yet")

        # TODO: Code generate this as a single expression
        def e_factored(mat, b, reverse=False):
            if reverse:
                r = reversed
            else:
                r = lambda i: i

            for num_root, den_root in zip(r(numroots), r(denroots)):
                f = lambdify((t, n0), multiply_vector((t - num_root)/(t -
                    den_root), n0), scipy_translations_autoeye,
                    printer=MatrixNumPyPrinter())
                b = f(mat, b)
            return float(p1q1)*b

        return e_factored



#! /usr/bin/env python
try:
    from setuptools import setup
    HAVE_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup
    HAVE_SETUPTOOLS = False

import numpy as np

VERSION = '1.0' # Also update __init__.py

setup_kwargs = {
    "version": VERSION,
    "description": 'Code geneartion tools for transmutation solvers',
    "license": 'BSD 3-clause',
    "author": 'ERGS',
    "author_email": 'ergsonomic@googlegroups.com',
    "url": 'https://github.com/ergs/transmutagen',
    "download_url": "https://github.com/ergs/transmutagen/zipball/" + VERSION,
    "classifiers": [
        "License :: OSI Approved",
        "Programming Language :: Python",
        ],
    "zip_safe": False,
    "data_files": [("", ['LICENSE', 'README.md']),],
    "package_data": {'transmutagen': ['data/gensolve.json', 'data/CRAM_cache/*']},
    }

if HAVE_SETUPTOOLS:
    setup_kwargs['install_requires'] = [
        'mpmath',
        'sympy',
        'matplotlib',
        'numpy',
        'scipy',
        'pyne',
        'jinja2',
        'gmpy2',
        'Cython'
        ]

if __name__ == '__main__':
    setup(
        name='transmutagen',
        packages=['transmutagen'],
        long_description=open('README.md').read(),
        include_dirs = [np.get_include()],
        **setup_kwargs
        )



"""
Regenerate the gensolve_json from ORIGEN with

    python -m transmutagen.generate_json /path/to/origen/libs/
"""

import os
import json
from argparse import ArgumentParser
from operator import itemgetter

from scipy.sparse import eye, csr_matrix
import numpy as np

from .tape9utils import normalize_tape9s, tape9_to_sparse

def csr_ij(mat):
    ij = {}
    for i, l, u in zip(range(mat.shape[0]), mat.indptr[:-1], mat.indptr[1:]):
        for p in range(l, u):
            ij[i, int(mat.indices[p])] = p
    return ij

def common_mat(mats):
    assert len({i.shape for i in mats}) == 1
    mats = [i.tocoo() for i in mats] + [eye(mats[0].shape[0], format='coo')]
    rows = np.hstack([i.row for i in mats])
    cols = np.hstack([i.col for i in mats])
    data = np.ones(len(rows))
    return csr_matrix((data, (rows, cols)))


def generate_json(tape9s, decaylib, outfile='transmutagen/data/gensolve.json'):
    mats, nucs = tape9_to_sparse(tape9s, phi=1.0, format='csr', decaylib=decaylib)
    mat = common_mat(mats)
    ij = csr_ij(mat)
    tofrom = [(nucs[j], nucs[i]) for i, j in sorted(ij, key=itemgetter(1))]
    os.makedirs(os.path.dirname(outfile), exist_ok=True)

    with open(outfile, 'w') as f:
        print("Writing", outfile)
        json.dump({
            'nucs': list(nucs),
            # JSON associative arrays can only have string keys
            'tofrom': tofrom,
            }, f, sort_keys=True, indent=4)

def main(args=None):
    p = ArgumentParser('generate_json', description="""Generate the JSON input
        file for gensolve from TAPE 9 files.""")
    p.add_argument('tape9s', nargs='+', help="""Paths to the TAPE9 files. If a
        path is a directory, a set of default libraries will be gathered from
        that directory (transmutagen.origen_all.ALL_LIBS)""")
    p.add_argument('-d', '--decay', help='path to the decay file, if needed',
                   default='decay.lib', dest='decaylib')
    p.add_argument('--outfile', default='transmutagen/data/gensolve.json',
        help="""File to save the JSON file to. The default is %(default)r.""")

    ns = p.parse_args(args=args)
    tape9s = normalize_tape9s(ns.tape9s)
    generate_json(tape9s, ns.decaylib, outfile=ns.outfile)


if __name__ == "__main__":
    main()
